import discord
import discord.utils
import enum
import logging
import re
import shlex
import asyncio
import json
from datetime import timedelta
from discord.ext import tasks
from math import floor, ceil
from functools import partial
from random import randint
from time import monotonic as time
from yt_dlp import YoutubeDL
from yt_dlp.utils import YoutubeDLError
from red_star.command_dispatcher import Command
from red_star.plugin_manager import BasePlugin
from red_star.rs_errors import UserPermissionError, CommandSyntaxError
from red_star.rs_utils import respond, split_message, find_user, is_positive

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Optional


class MusicPlayer(BasePlugin):
    name = "music_player"
    version = "2.2"
    author = "medeor413 (original by GTG3000)"
    description = "A plugin for playing audio from videos in a voice channel."
    default_config = {
        "max_queue_length": 30,
        "max_video_length": 1800,
        "vote_skip_threshold": 0.5,
        "print_queue_on_edit": True,
        "idle_disconnect_time": 300
    }
    default_global_config = {
        "save_audio": True,
        "video_cache_clear_age": 60 * 60 * 24 * 3,
        "cache_clear_interval": 60 * 60,
        "default_volume": 15,
        "youtube_dl_config": {
            "quiet": True,
            "restrictfilenames": True,
            "source_address": "0.0.0.0",
            "audioformat": "mp3",
            "default_search": "auto",
            "extractaudio": True,
            "logtostderr": False,
            "nocheckcertificate": True,
            "format": "bestaudio/best",
            "noplaylist": True,
            "no_color": True
        },
    }

    cache_reaper = None

    async def activate(self):
        self.downloaded_songs_folder = self.client.storage_dir / "music_cache"

        self.player: Optional[GuildPlayer] = None

        self.ydl_options = self.global_plugin_config["youtube_dl_config"].copy()
        self.ydl_options["logger"] = self.logger.getChild("youtube-dl")
        self.ydl_options["extract_flat"] = "in_playlist"
        self.ydl_options["outtmpl"] = str(self.downloaded_songs_folder /
                                          self.ydl_options.get("outtmpl", "%(id)s-%(extractor)s.%(ext)s"))

        self._port_old_storage()

        if not self.cache_reaper:
            self.cache_reaper = self.reap_cache
            self.cache_reaper.change_interval(seconds=self.global_plugin_config["cache_clear_interval"])
            self.cache_reaper.start()

    def _port_old_storage(self):
        old_storage_path = self.config_manager.config_path / "music_player.json"
        if old_storage_path.exists():
            with old_storage_path.open(encoding="utf-8") as fp:
                old_storage = json.load(fp)
            for guild_id, bans in old_storage.get("banned_users", {}).items():
                try:
                    new_storage = self.config_manager.storage_files[guild_id][self.name]
                except KeyError:
                    self.logger.warn(f"Server with ID {guild_id} not found! Is the bot still in this server?\n"
                                     f"Skipping conversion of this server's music ban storage...")
                    continue
                bans.update(new_storage.contents.get("banned_users", {}))
                new_storage.contents["banned_users"] = bans
                new_storage.save()
                new_storage.load()
            old_storage_path = old_storage_path.replace(old_storage_path.with_suffix(".json.old"))
            self.logger.info(f"Old music ban storage converted to new format. "
                             f"Old data now located at {old_storage_path} - you may delete this file.")

    async def deactivate(self):
        if self.player:
            self.player.stop()
            await self.player.voice_client.disconnect()

    async def on_voice_state_update(self, *_):
        # We don't care about the contents of the update, just that it happened. Check if the contents of our voice
        # channel have changed.
        if self.player:
            if len(self.player.voice_client.channel.members) <= 1:  # The bot is now alone.
                disconnect_time = (discord.utils.utcnow() + timedelta(
                        seconds=self.config["idle_disconnect_time"])).time()
                # If an interval is provided instead of an explicit time, the function will run immediately.
                self.player.idle_disconnect.change_interval(time=disconnect_time)
                self.player.idle_disconnect.start()
            elif self.player.idle_disconnect.is_running():  # The bot isn't alone - cancel the disconnect task.
                self.player.idle_disconnect.cancel()

    # Command functions

    @Command("JoinVoice", "JoinVC",
             doc="Tells the bot to join the voice channel you're currently in.",
             category="music_player")
    async def _join_voice(self, msg: discord.Message):
        try:
            voice_channel = msg.author.voice.channel
        except AttributeError:
            raise UserPermissionError("**ANALYSIS: User is not connected to voice channel.**")
        try:
            if self.player:
                await self.player.voice_client.move_to(voice_channel)
            else:
                self.player = await self.create_player(voice_channel, msg.channel)
        except asyncio.TimeoutError:
            raise CommandSyntaxError("Bot failed to join channel. Make sure bot can access the channel.")
        await respond(msg, f"**AFFIRMATIVE. Connected to channel {voice_channel.name}.**")
        return self.player

    @Command("LeaveVoice", "LeaveVC",
             doc="Tells the bot to leave the voice channel it's currently in.",
             category="music_player",
             optional_perms={"force_leave": {"mute_members"}})
    async def _leave_voice(self, msg: discord.Message):
        if not self.player:
            try:
                msg.guild.voice_client.stop()
                await msg.guild.voice_client.disconnect(force=True)
                await respond(msg, "**ANALYSIS: Disconnected from voice channel.**")
                return
            except AttributeError:
                await respond(msg, "**ANALYSIS: Bot is not currently in voice channel.**")
                return
        if self.player.is_playing \
                and not self._leave_voice.perms.check_optional_permissions("force_leave", msg.author, msg.channel):
            raise UserPermissionError
        self.player.queue.clear()
        self.player.already_played.clear()
        self.player.stop()
        await self.player.voice_client.disconnect()
        self.player = None
        await respond(msg, "**ANALYSIS: Disconnected from voice channel.**")

    @Command("PlaySong", "Play",
             doc="Tells the bot to queue video links for playing. User must be in the same channel as the bot. "
                 "Supports an arbitrary number of links in a single command.\n"
                 "Supports playlists. Additionally, subsets of playlists can be played using slice notation.\n"
                 "Subset syntax: https://example.url{1,3-5,-2,6-}. {1} selects the first video, {3-5} selects the "
                 "third through fifth videos, {-2} selects all videos up to the second, {6-} selects the sixth video "
                 "and all after it. Multiple slices can be used on one playlist by separating them with ,.",
             syntax="(url) [url_2] [url_3]...",
             category="music_player")
    async def _play_song(self, msg: discord.Message):
        if not self.player:
            self.player = await self._join_voice(msg)
        self.check_user_permission(msg.author, self.player)
        try:
            urls = shlex.split(msg.clean_content)[1:]
            # Is it supposed to be a search query, or a bunch of URLs?
            if len(urls) > 1:
                url_validator = re.compile(r"https?://(www\.)?[-a-zA-Z\d@:%._+~#=]{1,256}\.[a-zA-Z\d()]{1,6}"
                                           r"\b([-a-zA-Z\d()@:%_+.~#?&/=]*)")
                if any((not url_validator.fullmatch(url) and " " not in url) for url in urls):
                    # If it's just a plain search query (no URLs, no quoted spaces), join it back together for QoL
                    urls = [" ".join(urls)]
        except ValueError as e:
            raise CommandSyntaxError(e)
        await self.player.prepare_playlist(urls)

    @Command("SongQueue", "Queue",
             doc="Tells the bot to list the current song queue.",
             category="music_player")
    async def _list_queue(self, msg: discord.Message):
        if not self.player:
            await respond(msg, "**ANALYSIS: Bot is not in a voice channel.**")
            return
        elif self.player.queue and not self.player.current_song:
            await respond(msg, "**ANALYSIS: The queue is empty.**")
            return
        if self.player.current_song:
            await respond(msg, f"**ANALYSIS: Now playing:**\n```{self.player.print_now_playing()}```")
        if self.player.queue:
            queue_duration = pretty_duration(self.player.queue_duration)
            for split_msg in split_message(f"**ANALYSIS: Current queue: ({queue_duration})**"
                                           f"{self.player.print_queue()}"):
                await respond(msg, split_msg)

    @Command("NowPlaying", "SongInfo",
             doc="Displays detailed information about the currently playing song.",
             category="music_player")
    async def _song_info(self, msg: discord.Message):
        if not self.player or not self.player.current_song:
            await respond(msg, "**ANALYSIS: There is no song currently playing.**")
            return
        vid = self.player.current_song
        desc = vid.get("description", "*No description.*")
        if len(desc) > 2048:
            desc = desc[:2045] + "..."
        embed = discord.Embed(title=vid.get("title", "Unknown"), description=desc, url=vid["url"])
        if vid.get("thumbnail") is not None:
            embed.set_thumbnail(url=vid["thumbnail"])
        if vid.get("uploader") is not None:
            embed.set_author(name=vid["uploader"])
        if vid.get("tags") is not None:
            embed.set_footer(text=f"Tags: {', '.join(vid['tags'])}")
        play_time, duration, _ = self.player.progress
        rating_field = f"Views: {vid.get('view_count', 'Unknown'):,}."
        if vid.get("like_count") is not None:
            rating_field += f" {vid['like_count']:,}👍"
        if vid.get("dislike_count") is not None:
            rating_field += f"/{vid['dislike_count']:,}👎"
        if vid.get("average_rating") is not None:
            rating_field += f" (Average rating: {vid['average_rating']:.2f})"
        embed.add_field(name="Ratings", value=rating_field)
        await respond(msg, embed=embed)

    @Command("DeleteSong", "DelSong", "RMSong",
             doc="Removes a song from the bot's queue.",
             syntax="(index)",
             perms="mute_members",
             category="music_player")
    async def _delete_song(self, msg: discord.Message):
        if not self.player:
            await respond(msg, "**ANALYSIS: Bot is not in a voice channel.**")
            return
        elif not self.player.queue:
            await respond(msg, "**ANALYSIS: The queue is empty.**")
            return
        try:
            index = int(msg.clean_content.split(None, 1)[1])
        except IndexError:
            raise CommandSyntaxError("No arguments provided")
        except ValueError:
            raise CommandSyntaxError(f"{msg.clean_content.split(None, 1)[1]} is not a valid integer.")
        if index == 0:
            raise CommandSyntaxError("Use Skip to skip the currently playing song")
        elif index < 0:
            raise CommandSyntaxError("Provided integer must be positive")
        index -= 1
        try:
            del_song = self.player.queue.pop(index)
            if index in self.player.already_played:
                self.player.already_played.remove(index)
            new_already_played = set()
            for i in self.player.already_played:
                if i > index:
                    i -= 1
                new_already_played.add(i)
            self.player.already_played = new_already_played
        except IndexError:
            raise CommandSyntaxError("Integer provided is not a valid index")
        await respond(msg, f"**AFFIRMATIVE. Deleted song at position {index + 1} ({del_song['title']}).**")
        if self.config["print_queue_on_edit"] and self.player.queue:
            for split_msg in split_message(f"**ANALYSIS: Current queue:**{self.player.print_queue()}"):
                await respond(msg, split_msg)

    @Command("SongVolume", "Volume",
             doc="Sets the volume at which the player plays music, between 0 and 100.\nUse --temporary flag to "
                 "automatically reset volume at the end of the current song.",
             syntax="[-t/--temporary] (0-100)",
             category="music_player")
    async def _set_volume(self, msg: discord.Message):
        self.check_user_permission(msg.author, self.player)
        new_volume = 0
        try:
            temp_flag = False
            new_volume = msg.clean_content.split(None, 2)[1:]
            if len(new_volume) > 1:
                if new_volume[0] in ("-t", "--temporary"):
                    temp_flag = True
                    new_volume = int(new_volume[1])
                else:
                    raise CommandSyntaxError
            else:
                new_volume = int(new_volume[0])
            if new_volume == 0:
                raise CommandSyntaxError("Please use PauseSong to mute the bot")
            elif not 0 < new_volume <= 100:
                raise CommandSyntaxError("Expected value between 0 and 100")
        except IndexError:
            await respond(msg, f"**ANALYSIS: Current volume is {int(self.player.volume * 100)}%.**")
            return
        except ValueError:
            raise CommandSyntaxError(f"Value {new_volume} is not a valid integer")
        if temp_flag:
            self.player.prev_volume = self.player.volume
        self.player.volume = new_volume / 100
        await respond(msg, f"**AFFIRMATIVE. Set volume to {new_volume}%.**")

    @Command("SongMode",
             doc="Tells the bot in what order the bot should play its queue, or prints the current mode if no mode "
                 "is specified.",
             syntax="[n/normal|rs/repeat_song|rq/repeat_queue|s/shuffle|sr/shuffle_repeat]",
             category="music_player")
    async def _song_mode(self, msg: discord.Message):
        self.check_user_permission(msg.author, self.player)
        try:
            arg = msg.clean_content.split(None, 1)[1].lower()
        except IndexError:
            await respond(msg, f"**ANALYSIS: Current song mode is {self.player.song_mode}.**")
            return
        if arg in ("n", "normal", "reset"):
            self.player.song_mode = SongMode.NORMAL
        elif arg in ("rs", "repeat_song", "loop"):
            self.player.song_mode = SongMode.REPEAT_SONG
        elif arg in ("rq", "repeat", "repeat_queue", "cycle"):
            self.player.song_mode = SongMode.REPEAT_QUEUE
        elif arg in ("s", "shuffle", "random"):
            self.player.song_mode = SongMode.SHUFFLE
        elif arg in ("sr", "shuffle_repeat", "random_repeat"):
            self.player.song_mode = SongMode.SHUFFLE_REPEAT
            self.player.already_played.clear()
        else:
            raise CommandSyntaxError(f"Argument {arg} is not a valid mode")
        await respond(msg, f"**AFFIRMATIVE. Song mode changed to {self.player.song_mode}.**")

    @Command("Shuffle",
             doc="Enables or disables shuffling of song order.",
             syntax="[true/yes/on/false/no/off",
             category="music_player")
    async def _shuffle_toggle(self, msg: discord.Message):
        self.check_user_permission(msg.author, self.player)
        shuffle_on = self.player.song_mode in (SongMode.SHUFFLE, SongMode.SHUFFLE_REPEAT)
        try:
            arg = is_positive(msg.clean_content.split(None, 1)[1])
        except IndexError:
            await respond(msg, f"**ANALYSIS: Shuffle is currently {'enabled' if shuffle_on else 'disabled'}.**")
            return
        if arg:
            if shuffle_on:
                await respond(msg, f"**ANALYSIS: Shuffle is already enabled.**")
            else:
                if self.player.song_mode is SongMode.NORMAL:
                    self.player.song_mode = SongMode.SHUFFLE
                    await respond(msg, f"**AFFIRMATIVE. Shuffle enabled.**")
                elif self.player.song_mode is SongMode.REPEAT_QUEUE:
                    self.player.song_mode = SongMode.SHUFFLE_REPEAT
                    await respond(msg, f"**AFFIRMATIVE. Shuffle enabled with repeat.**")
                elif self.player.song_mode is SongMode.REPEAT_SONG:
                    await respond(msg, "**NEGATIVE. Shuffle cannot be enabled at the same time as song repeat.**")
        else:
            if shuffle_on:
                if self.player.song_mode is SongMode.SHUFFLE:
                    self.player.song_mode = SongMode.NORMAL
                    await respond(msg, f"**AFFIRMATIVE. Shuffle disabled.**")
                elif self.player.song_mode is SongMode.SHUFFLE_REPEAT:
                    self.player.already_played.clear()
                    self.player.song_mode = SongMode.REPEAT_QUEUE
                    await respond(msg, f"**AFFIRMATIVE. Shuffle disabled. Queue repeat still enabled.**")
            else:
                await respond(msg, "**ANALYSIS: Shuffle is not enabled.**")

    @Command("Repeat",
             doc="Modifies repeating mode of song queue.",
             syntax="[song/one/queue/all/false/no/off]",
             category="music_player")
    async def _repeat_mode(self, msg: discord.Message):
        self.check_user_permission(msg.author, self.player)
        try:
            arg = msg.clean_content.split(None, 1)[1].lower()
            if arg not in ("song", "one", "queue", "all"):
                arg = is_positive(arg)
                if arg:
                    raise CommandSyntaxError
        except IndexError:
            if self.player.song_mode is SongMode.REPEAT_SONG:
                repeat_mode = "set to repeat current song"
            elif self.player.song_mode is SongMode.REPEAT_QUEUE:
                repeat_mode = "set to repeat the queue"
            elif self.player.song_mode is SongMode.SHUFFLE_REPEAT:
                repeat_mode = "set to repeat the queue. Shuffle is also enabled"
            else:
                repeat_mode = "disabled"
            await respond(msg, f"**ANALYSIS: Repeat is currently {repeat_mode}.**")
            return
        if arg in ("song", "one"):
            if self.player.song_mode is SongMode.REPEAT_SONG:
                await respond(msg, "**ANALYSIS: Song repeat is already enabled.**")
            elif self.player.song_mode in (SongMode.SHUFFLE, SongMode.SHUFFLE_REPEAT):
                await respond(msg, "**NEGATIVE. Song repeat cannot be enabled at the same time as shuffle.**")
            else:
                self.player.song_mode = SongMode.REPEAT_SONG
                await respond(msg, "**AFFIRMATIVE. Repeat mode set to song.**")
        elif arg in ("queue", "all"):
            if self.player.song_mode in (SongMode.REPEAT_QUEUE, SongMode.SHUFFLE_REPEAT):
                await respond(msg, "**ANALYSIS: Queue repeat is already enabled.**")
            elif self.player.song_mode is SongMode.SHUFFLE:
                self.player.song_mode = SongMode.SHUFFLE_REPEAT
                await respond(msg, "**AFFIRMATIVE. Repeat mode set to queue with shuffle.**")
            else:
                self.player.song_mode = SongMode.REPEAT_QUEUE
                await respond(msg, "**AFFIRMATIVE. Repeat mode set to queue.**")
        else:
            if self.player.song_mode in (SongMode.NORMAL, SongMode.SHUFFLE):
                await respond(msg, "**ANALYSIS: Repeat mode is disabled.**")
            elif self.player.song_mode is SongMode.SHUFFLE_REPEAT:
                self.player.already_played.clear()
                self.player.song_mode = SongMode.SHUFFLE
                await respond(msg, "**AFFIRMATIVE. Repeat mode disabled. Shuffle still enabled.**")
            else:
                self.player.song_mode = SongMode.NORMAL
                await respond(msg, "**AFFIRMATIVE. Repeat mode disabled.**")

    @Command("SkipSong", "Skip",
             doc="Tells the bot to skip the currently playing song.",
             category="music_player",
             optional_perms={"force_skip": {"mute_members"}})
    async def _skip_song(self, msg: discord.Message):
        self.check_user_permission(msg.author, self.player)
        if not self.player or not self.player.is_playing:
            await respond(msg, "**ANALYSIS: No music currently playing.**")
            return
        elif self._skip_song.perms.check_optional_permissions("force_skip", msg.author, msg.channel):
            await respond(msg, "**ANALYSIS: Skipping to next song in queue...**")
            self.player.stop()
        else:
            await self.player.skip_vote(msg.author.id)

    @Command("PauseSong", "Pause", "ResumeSong", "Resume",
             doc="Tells the bot to pause or resume the current song.",
             perms="mute_members",
             category="music_player")
    async def _pause_song(self, msg: discord.Message):
        if not self.player or self.player.is_playing is False:
            await respond(msg, "**WARNING: No music currently playing.**")
            return
        if self.player.toggle_pause():
            await respond(msg, "**ANALYSIS: Song paused.**")
        else:
            await respond(msg, "**ANALYSIS: Song resumed.**")

    @Command("StopMusic", "StopSong", "Stop",
             doc="Tells the bot to stop playing songs and empty the queue.",
             perms="mute_members",
             category="music_player")
    async def _stop_music(self, msg: discord.Message):
        if not self.player or not self.player.is_playing:
            await respond(msg, "**ANALYSIS: No music currently playing.")
            return
        self.player.queue.clear()
        self.player.already_played.clear()
        self.player.stop()
        await respond(msg, "**ANALYSIS: The music has been stopped and the queue has been cleared.**")

    @Command("MusicBan",
             doc="Bans or unbans a user from using music bot commands, or lists bans if no user specified..",
             syntax="(user)",
             perms="mute_members",
             category="music_player")
    async def _music_ban(self, msg: discord.Message):
        try:
            ban_store = self.storage["banned_users"]
        except KeyError:
            ban_store = self.storage["banned_users"] = []
        try:
            user = find_user(msg.guild, msg.content.split(None, 1)[1])
        except IndexError:
            user_li = [find_user(msg.guild, x) for x in ban_store]
            banned_list = "\n".join(str(x) if x else "Unknown" for x in user_li)
            if not banned_list:
                banned_list = "None."
            for split_msg in split_message(f"**ANALYSIS: Banned users:**\n```\n{banned_list}```"):
                await respond(msg, split_msg)
            return
        if user:
            if user.id in ban_store:
                ban_store.remove(user.id)
                await respond(msg, f"**AFFIRMATIVE. User {user} has been unbanned from using the music module.**")
                return
            else:
                ban_store.append(user.id)
            await respond(msg, f"**AFFIRMATIVE. User {user} has been banned from using the music module.**")
            self.storage_file.save()
        else:
            raise CommandSyntaxError(f"No such user {msg.clean_content.split(None, 1)[1]}")

    @Command("MusicConfig",
             doc="Allows the configuration of per-server music bot options.\n"
                 "Valid options: 'max_queue_length', 'max_video_length', 'vote_skip_threshold', "
                 "'print_queue_on_edit', 'idle_disconnect_time'",
             syntax="[(option) (value)]",
             perms="manage_guild",
             category="music_player")
    async def _music_config(self, msg: discord.Message):
        try:
            opt, val = msg.clean_content.split(None, 2)[1:]
        except ValueError:
            current_conf = "\n".join(f"{k}: {v}" for k, v in self.config.items())
            await respond(msg, f"**ANALYSIS: Current configuration:**```{current_conf}```")
            return
        opt = opt.lower()
        if opt in ("max_queue_length", "max_video_length", "idle_disconnect_time"):
            try:
                val = int(val)
            except ValueError:
                raise CommandSyntaxError(f"{val} is not a valid integer")
        elif opt == "vote_skip_threshold":
            try:
                val = float(val)
                if not 0 < val <= 1:
                    raise ValueError
            except ValueError:
                raise CommandSyntaxError("Value must be a number between 0 and 1.")
        elif opt == "print_queue_on_edit":
            val = is_positive(val)
        else:
            raise CommandSyntaxError(f"Option {opt} does not exist")
        self.config[opt] = val
        await respond(msg, f"**AFFIRMATIVE. Option `{opt}` edited to `{val}` successfully.**")

    # Utility functions

    @tasks.loop(hours=1)
    async def reap_cache(self):
        now = time()
        for file in self.downloaded_songs_folder.iterdir():
            if now - file.stat().st_atime > self.global_plugin_config["video_cache_clear_age"]:
                try:
                    file.unlink()
                    self.logger.debug(f"Deleted old video cache file {file}.")
                except OSError:
                    continue

    async def create_player(self, voice_channel: discord.VoiceChannel, text_channel: discord.TextChannel):
        async with text_channel.typing():
            voice_client = await voice_channel.connect()
            player = GuildPlayer(self, voice_client, text_channel)
            self.players = player
        return player

    def check_user_permission(self, user: discord.Member, player: "GuildPlayer"):
        try:
            if user.id in self.storage["banned_users"]:
                raise UserPermissionError("You are banned from using the music player.")
            if user not in player.voice_client.channel.members:
                raise UserPermissionError("You are not in the voice channel.")
        except AttributeError:
            raise UserPermissionError("The bot is not currently in a voice channel.")
        return True


class GuildPlayer:
    def __init__(self, parent: MusicPlayer, voice_client: discord.VoiceClient, channel: discord.TextChannel):
        self.parent = parent
        self.text_channel = channel
        self.voice_client = voice_client
        self.logger = logging.getLogger(f"red_star.plugin.music_player.{self.voice_client.guild.id}.player")
        self.queue = []
        self.already_played = set()
        self.is_playing = False
        self.current_song = {}
        self.song_mode = SongMode.NORMAL
        self.prev_volume = None
        self._volume = self.parent.global_plugin_config["default_volume"] / 100
        self._song_start_time = None
        self._song_pause_time = None
        self._skip_votes = set()

    async def prepare_playlist(self, urls: [str]):
        async with self.text_channel.typing():
            # Fetch video info
            with YoutubeDL(self.parent.ydl_options) as ydl:
                # Create a list of videos to be queued
                to_queue = []
                for url in urls:
                    try:
                        pl_slice = re.match(r"([^{`]+)`*(?:{([^}]*)})?", url)
                        vid_info = await asyncio.get_running_loop().run_in_executor(
                                None, partial(ydl.extract_info, pl_slice[1], download=False))
                    except YoutubeDLError as e:
                        await self.text_channel.send(f"**WARNING. An error occurred while downloading video <{url}>. "
                                                     f"It will not be queued.\nError details:** `{e}`")
                        continue
                    # If it's a playlist, we need to flatten it into our to_queue list, and deal with the slicing
                    if vid_info.get("_type") == "playlist":
                        # Slice handling
                        if pl_slice[2]:
                            pl_slices = pl_slice[2].split(",")
                            for s in pl_slices:
                                # Ranges handling
                                if "-" in s:
                                    s_start, s_end = s.split("-", 1)
                                    try:
                                        s_start = int(s_start) - 1
                                    except ValueError:
                                        s_start = 0
                                    try:
                                        s_end = int(s_end)
                                        if s_end > len(vid_info["entries"]):
                                            raise ValueError
                                    except ValueError:
                                        s_end = len(vid_info["entries"])
                                    if s_end < s_start:
                                        continue
                                    to_queue.extend(vid_info["entries"][s_start:s_end])
                                # Single values
                                else:
                                    try:
                                        s = int(s) - 1
                                        to_queue.append(vid_info["entries"][s])
                                    except (ValueError, IndexError):
                                        continue
                        # No slice
                        else:
                            to_queue.extend(vid_info["entries"])
                    # If it's just a video, throw it in
                    else:
                        to_queue.append(vid_info)
            # Queuing is handled by _enqueue_playlist function. Don't bother it if we got nothing.
            if len(to_queue) > 0:
                asyncio.create_task(self._enqueue_playlist(to_queue))
                return

    async def _enqueue_playlist(self, entries: [dict]):
        if len(self.queue) > 0:
            time_until_song = self.queue_duration + (self.current_song.get("duration", 0) - self.play_time)
            time_until_song = f"\nTime until your song: {pretty_duration(time_until_song)}"
            await self.text_channel.send(f"**ANALYSIS: Attempting to queue {len(entries)} videos. Your playback will "
                                         f"begin shortly.{time_until_song}**")
        orig_len = len(self.queue)
        async with self.text_channel.typing():
            with YoutubeDL(self.parent.ydl_options) as ydl:
                for vid in entries:
                    # We only want to extract info if we don't already have it. Things get a little funky otherwise.
                    if vid.get("_type") in ("url", "url_transparent"):
                        try:
                            vid = await asyncio.get_running_loop().run_in_executor(
                                    None, partial(ydl.extract_info, vid["url"], download=False))
                        # Skip broken videos while trying the rest
                        except YoutubeDLError as e:
                            await self.text_channel.send(f"**WARNING. An error occurred while downloading video "
                                                         f"<{vid['url']}>. It will not be queued.\nError details:** "
                                                         f"`{e}`")
                            continue
                    # Abort once the queue is full
                    if len(self.queue) >= self.parent.config["max_queue_length"]:
                        await self.text_channel.send(f"**WARNING: The queue is full. No more videos will be added.**")
                        break
                    # Skip over videos that are too long
                    elif vid.get("duration", 0) > self.parent.config["max_video_length"]:
                        max_len = pretty_duration(self.parent.config["max_video_length"])
                        await self.text_channel.send(f"**WARNING: Video {vid['title']} exceeds the maximum video"
                                                     f" length ({max_len}). It will not be added.**")
                        continue
                    try:
                        await self._process_video(vid)
                    except TypeError:
                        continue
                    if not self.is_playing:
                        try:
                            asyncio.create_task(self._play())
                        except discord.ClientException:
                            pass
        await self.text_channel.send(f"**ANALYSIS: Queued {len(self.queue) - orig_len} videos.**")
        if self.parent.config["print_queue_on_edit"] and self.queue:
            final_msg = f"**ANALYSIS: Current queue:**{self.print_queue()}"
            for msg in split_message(final_msg):
                await self.text_channel.send(msg)

    async def _process_video(self, vid: dict):
        if not vid:
            raise TypeError
        if self.parent.global_plugin_config["save_audio"] and not vid.get("is_live", False):
            with YoutubeDL(self.parent.ydl_options) as ydl:
                try:
                    await asyncio.get_running_loop().run_in_executor(None, partial(ydl.process_info, vid))
                    vid["filename"] = ydl.prepare_filename(vid)
                except YoutubeDLError as e:
                    await self.text_channel.send(f"**WARNING. An error occurred while downloading video "
                                                 f"{vid['title']}. It will not be queued.\nError details:** `{e}`")
                    return
        vid.setdefault("title", "Unknown")
        vid.setdefault("is_live", False)
        vid.setdefault("duration", 0)
        if vid["duration"] < 0:
            vid["duration"] = 0
        self.queue.append(vid)

    async def _play(self):
        try:
            if self.song_mode == SongMode.SHUFFLE_REPEAT:
                try:
                    song_indices = {i for i, _ in enumerate(self.queue)}
                    eligible_songs = song_indices - self.already_played
                    if not eligible_songs:
                        self.already_played.clear()
                        eligible_songs = song_indices
                    eligible_songs = tuple(eligible_songs)
                    next_song = eligible_songs[randint(0, len(eligible_songs) - 1)]
                    self.already_played.add(next_song)
                    next_song = self.queue[next_song]
                except ValueError:
                    next_song = self.queue[0]
            elif self.song_mode == SongMode.SHUFFLE:
                try:
                    next_song = self.queue.pop(randint(0, len(self.queue) - 1))
                except ValueError:
                    next_song = self.queue.pop()
            elif self.song_mode == SongMode.REPEAT_QUEUE:
                if self.current_song:
                    self.queue.append(self.current_song)
                next_song = self.queue.pop(0)
            elif self.song_mode == SongMode.REPEAT_SONG:
                if self.current_song:
                    next_song = self.current_song
                else:
                    next_song = self.queue.pop(0)
            else:
                next_song = self.queue.pop(0)
        except IndexError:
            await self.text_channel.send("**ANALYSIS: Queue complete.**")
            self.is_playing = False
            self.current_song = {}
            self._skip_votes = set()
            self.voice_client.stop()
            try:
                self.voice_client.source.cleanup()
            except AttributeError:
                pass
            return
        before_args = ""
        if self.parent.global_plugin_config["save_audio"] and not next_song["is_live"]:
            file = next_song["filename"]
        else:
            file = next_song["url"]
            before_args += " -reconnect 1 -reconnect_streamed 1 -reconnect_delay_max 30"
        self.is_playing = True
        if self.prev_volume:
            self.volume = self.prev_volume
            self.prev_volume = None
            await self.text_channel.send(f"**ANALYSIS: Volume reset to {int(self.volume * 100)}%.**")
        source = discord.PCMVolumeTransformer(discord.FFmpegPCMAudio(file, before_options=before_args, options="-vn"),
                                              volume=self._volume)
        self.voice_client.play(source, after=partial(self._after, loop=asyncio.get_running_loop()))
        self._song_start_time = time()
        self.current_song = next_song
        self._skip_votes = set()
        await self.text_channel.send(f"**NOW PLAYING: {next_song['title']}.**")

    def toggle_pause(self) -> bool:
        if self.voice_client.is_paused():
            self.voice_client.resume()
            self._song_start_time += (time() - self._song_pause_time)
            return False
        else:
            self.voice_client.pause()
            self._song_pause_time = time()
            return True

    def _after(self, error: Exception, loop: asyncio.AbstractEventLoop):
        if error:
            self.logger.error(error)
        asyncio.run_coroutine_threadsafe(self._play(), loop)

    def stop(self):
        self.is_playing = False
        self._skip_votes = set()
        try:
            self.voice_client.source.cleanup()
        except AttributeError:
            pass
        self.voice_client.stop()

    @property
    def play_time(self) -> float:
        try:
            if self.voice_client.is_paused():
                return self._song_pause_time - self._song_start_time
            else:
                return time() - self._song_start_time
        except TypeError:
            return 0

    @property
    def progress(self) -> (float, float, float):
        try:
            fraction = self.play_time / self.current_song["duration"]
        except ZeroDivisionError:
            fraction = 0
        return self.play_time, self.current_song["duration"], fraction

    async def skip_vote(self, uid):
        if uid not in self._skip_votes:
            self._skip_votes.add(uid)
        else:
            await self.text_channel.send("**NEGATIVE. Skip vote already recorded.**")
            return
        total_users = len(self.voice_client.channel.members) - 1  # Don't want to count the bot itself
        threshold = self.parent.config["print_queue_on_edit"]
        vote_count = len(self._skip_votes)
        if vote_count / total_users >= threshold:
            await self.text_channel.send("**AFFIRMATIVE. Skipping current song.**")
            self.stop()
        else:
            votes_needed = ceil(total_users * threshold) - vote_count
            await self.text_channel.send(f"**AFFIRMATIVE. Skip vote recorded. {votes_needed} votes needed to skip.**")

    @property
    def volume(self) -> float:
        return self._volume

    @volume.setter
    def volume(self, val: float):
        self._volume = val
        try:
            self.voice_client.source.volume = val
        except AttributeError:
            pass

    @property
    def queue_duration(self) -> float:
        try:
            return sum(vid["duration"] for vid in self.queue)
        except TypeError:
            return 0

    def print_queue(self) -> str:
        str_list = []
        for i, vid in enumerate(self.queue):
            duration = fixed_width_duration(vid.get("duration", 0))
            # Title truncating and padding
            title = vid.get("title", "Unknown")
            if len(title) >= 59:
                title = title[:56] + "..."
            str_list.append(f"[{i+1:02d}][{title:-<59}][{duration}]")
        return "```\n{}```".format("\n".join(str_list))

    def print_now_playing(self) -> str:
        play_time, duration, progress = self.progress
        dur_str = fixed_width_duration(play_time) + "/" + fixed_width_duration(duration)
        progbar = progress_bar(progress, 70)
        title = self.current_song.get("title", "Unknown")
        if len(title) > 57:
            title = title[:54] + "..."
        return f"[{title:-<57}][{dur_str}]\n[{progbar}]"

    @tasks.loop(minutes=5, count=1)
    async def idle_disconnect(self):
        self.stop()
        await self.voice_client.disconnect()
        self.parent.player = None


def seconds_to_minutes(secs: float, hours: bool = False) -> tuple[int, int] | tuple[int, int, int]:
    mn, sec = divmod(secs, 60)
    if hours:
        hr, mn = divmod(mn, 60)
        return int(hr), int(mn), int(ceil(sec))
    else:
        return int(mn), int(ceil(sec))


def pretty_duration(seconds: float) -> str:
    if seconds <= 0:
        return "Unknown"
    elif seconds >= 3600:
        hours, minutes, seconds = seconds_to_minutes(seconds, hours=True)
        return f"{hours}:{minutes:02d}:{seconds:02d}"
    else:
        minutes, seconds = seconds_to_minutes(seconds)
        return f"{minutes:02d}:{seconds:02d}"


def fixed_width_duration(seconds: float) -> str:
    minutes, seconds = seconds_to_minutes(seconds)
    if seconds <= 0:
        return "--:--"
    if minutes > 99:
        if seconds >= 30:
            minutes += 1
        return f"{minutes:04d}m"
    else:
        return f"{minutes:02d}:{seconds:02d}"


def progress_bar(progress: float, length: int = 70) -> str:
    if not 0 <= progress <= 1:
        raise ValueError("First argument is expected to be a floating-point between 0 and 1.")
    bars = floor(length * progress)
    return "█" * bars + "-" * (length - bars)


class SongMode(enum.Enum):
    NORMAL = "normal"
    REPEAT_SONG = "repeat_song"
    REPEAT_QUEUE = "repeat_queue"
    SHUFFLE = "shuffle"
    SHUFFLE_REPEAT = "shuffle_repeat"

    def __str__(self) -> str:
        return self.name



import discord
from asyncio import sleep
from string import capwords
from red_star.plugin_manager import BasePlugin
from red_star.rs_errors import UserPermissionError
from red_star.rs_utils import respond
from red_star.rs_version import version
from red_star.command_dispatcher import Command

# Yes, I know about multiline strings. That breaks the max line length rule.
BASIC_HELP_TEXT = (
    "```\n"
    "This command can be used to get more information on the various commands the bot has to offer. Type {deco}Help "
    "categories to get a list of command categories. Then, type {deco}Help followed by one of the category names "
    "listed to get a list of commands in that category, and type {deco}Help followed by a command name for detailed "
    "information on that command.\n\n"

    "Some basic usage information:\n"
    "- In command syntax, a parameter in (parentheses) is required, and a parameter in [brackets] is optional. "
    "Don't type the parentheses or brackets!\n"
    "- If you see a parameter written like [-o/--option], this is a flag. You can use either -o or --option to "
    "specify it. If you see this written [-o/--option something], then the flag needs a parameter as well: "
    "[-n/--name name] might be used by typing \"-n John\", for example.\n"
    "- If you need to write a parameter with spaces, say a user named \"Some Thing\", you should put quotes around "
    "it in the command. Single or double quotes will work. To type quotes inside a quoted name, type a \\ before the "
    "quote, ex. \"Some \\\"Thing\\\"\".\n"
    "- If a command asks for a user, you can specify the user by their username, server nickname, username with "
    "discriminator, user ID, or by mentioning them. Be specific - if two users have the same name, the bot will pick "
    "one of them, and it might not be the one you wanted!\n"
    "- If a command asks for a role, you can similarly specify it using the role's name or ID, or by mentioning it.\n"
    "- Several commands ask for a yes/no type input. Many words will work for these, including true, yes, on, "
    "and enable for yes, and false, no, off, and disable for no."
    "```"
)


class Info(BasePlugin):
    name = "info"
    version = "1.1"
    author = "medeor413"
    description = "A plugin that provides commands for fetching information about other commands, or the bot itself."

    async def activate(self):
        self.commands = {}
        self.categories = {}

    async def on_all_plugins_loaded(self):
        await sleep(1)
        await self.build_help()

    async def on_plugin_activated(self, _):
        await sleep(1)
        await self.build_help()

    async def on_plugin_deactivated(self, _):
        await sleep(1)
        await self.build_help()

    async def build_help(self):
        self.commands = self.command_dispatcher.commands
        self.categories = {}
        for command in self.commands.values():
            name = command.name
            cmd_category = command.category.lower()
            if cmd_category not in self.categories:
                self.categories[cmd_category] = {}
            self.categories[cmd_category][name] = command

    @Command("Help",
             doc="Displays information on commands.",
             syntax="[category/command]",
             category="info")
    async def _help(self, msg: discord.Message):
        if not self.categories:
            await self.build_help()
        try:
            search = msg.clean_content.split(None, 1)[1].lower()
        except IndexError:  # No category or command specified. Give some general help text.
            deco = self.command_dispatcher.config["command_prefix"]
            await respond(msg, BASIC_HELP_TEXT.format(deco=deco))
            return
        if search == "categories":
            categories = "\n".join(sorted([capwords(x, "_") for x in self.categories.keys()]))
            await respond(msg, f"**ANALYSIS: Command categories:**```\n{categories}\n```")
        elif search in self.categories.keys():
            name = capwords(search, "_")
            cmds = {x.name for x in self.categories[search].values()
                    if x.perms.check_permissions(msg.author, msg.channel)}
            cmds = sorted(list(cmds))
            if cmds:
                text = "\n".join(cmds)
                await respond(msg, f"**ANALYSIS: Category {name}:**```\n{text}\n```")
            else:
                await respond(msg, "**WARNING: You do not have permission for any command in this category.**")
        elif search in [x.lower() for x in self.commands.keys()]:
            cmd = self.commands[search]
            name = cmd.name
            syntax = cmd.syntax
            if not syntax:
                syntax = "N/A"
            doc = cmd.__doc__
            perms = cmd.perms
            category = capwords(cmd.category, "_")
            aliases = f"(Aliases: {', '.join(cmd.aliases)})" if cmd.aliases else ""
            if not perms.check_permissions(msg.author, msg.channel):
                raise UserPermissionError
            text = f"**ANALYSIS: Command {name}:**```\n{name} (Category {category}) {aliases}\n\n{doc}\n\n" \
                   f"Syntax: {syntax}\n```"
            await respond(msg, text)
        else:
            await respond(msg, f"**WARNING: No such category or command {search}**")

    @Command("About",
             doc="Displays information about the bot.",
             category="info")
    async def _about(self, msg: discord.Message):
        deco = self.command_dispatcher.config["command_prefix"]
        desc = f"Red Star: General purpose command AI for Discord.\n" \
               f"Use {deco}help for command information."
        em = discord.Embed(title="About Red Star", color=0xFF0000, description=desc)
        em.set_thumbnail(url="https://raw.githubusercontent.com/medeor413/Red_Star/master/default_avatar.png")
        em.add_field(name="GitHub", value="https://github.com/medeor413/Red_Star")
        em.add_field(name="Version", value=version)
        await respond(msg, embed=em)



import discord
import shlex
from red_star.plugin_manager import BasePlugin
from red_star.command_dispatcher import Command, CommandPermissions
from red_star.rs_errors import CommandSyntaxError, UserPermissionError
from red_star.rs_utils import respond, find_user, find_role

# TODO: Refactors. This could probably be deduplicated a lot. Also evaluate whether update() method and creation of
# new permission objects is actually necessary...


class CommandPermissionCommands(BasePlugin):
    name = "command_permission_commands"
    version = "1.0"
    author = "medeor413"
    description = "A somewhat confusingly-named plugin that implements commands to edit the necessary permissions " \
                  "for other commands."

    async def activate(self):
        self.permission_overrides = self.command_dispatcher.config["permission_overrides"]
        self.commands = self.command_dispatcher.commands

    @Command("AddCommandPermission", "AddCmdPerm",
             doc="Adds a required permission(s) to a given command. All of these permissions are required to run a "
                 "command.",
             syntax="(command) (permissions_to_add)",
             perms="manage_guild",
             category="command_permissions")
    async def _add_command_permission(self, msg: discord.Message):
        command_to_edit, perms_to_add = self._validate_arguments_permissions(msg.clean_content)
        existing_perms = command_to_edit.perms
        successfully_added_perms = {x.upper() for x in perms_to_add if x not in existing_perms.permissions_all}

        if not existing_perms.check_permissions(msg.author, msg.channel):
            raise UserPermissionError

        new_perms = CommandPermissions.from_existing(existing_perms,
                                                     permissions_all=existing_perms.permissions_all | perms_to_add)

        if not new_perms.check_permissions(msg.author, msg.channel):
            raise UserPermissionError

        command_to_edit.perms.update(new_perms)
        self.permission_overrides[command_to_edit.name.lower()]["permissions_all"] = list(new_perms.permissions_all)
        self.config_manager.save_config()

        await respond(msg, f"**ANALYSIS: Command {command_to_edit.name} edited successfully. Added required "
                           f"permissions:** `{', '.join(successfully_added_perms)}`")

    @Command("RemoveCommandPermission", "RmCmdPerm",
             doc="Removes a required permission(s) from a given command. If all required permissions are removed "
                 "from a command, and it has no \"any of\" permissions, it can be used by anyone.",
             syntax="(command) (permissions_to_remove)",
             perms="manage_guild",
             category="command_permissions")
    async def _rm_command_permission(self, msg: discord.Message):
        command_to_edit, perms_to_remove = self._validate_arguments_permissions(msg.clean_content)
        existing_perms = command_to_edit.perms

        if not existing_perms.check_permissions(msg.author, msg.channel):
            raise UserPermissionError

        new_perms = CommandPermissions.from_existing(existing_perms,
                                                     permissions_all=existing_perms.permissions_all - perms_to_remove)

        difference_set = existing_perms.permissions_all - new_perms.permissions_all

        command_to_edit.perms.update(new_perms)
        self.permission_overrides[command_to_edit.name.lower()]["permissions_all"] = list(new_perms.permissions_all)
        self.config_manager.save_config()

        await respond(msg, f"**ANALYSIS: Command {command_to_edit.name} edited successfully. Removed required "
                           f"permissions:** `{', '.join(x.upper() for x in difference_set)}`")

    @Command("AddCommandPermissionAnyOf", "AddCmdPermAny",
             doc="Adds an \"any of\" permission to a given command. Any one of the permissions in the \"any of\" "
                 "list are required to run a command.",
             syntax="(command) (permissions_to_add)",
             perms="manage_guild",
             category="command_permissions")
    async def _add_command_permission_any(self, msg: discord.Message):
        command_to_edit, perms_to_add = self._validate_arguments_permissions(msg.clean_content)
        existing_perms = command_to_edit.perms
        successfully_added_perms = {x.upper() for x in perms_to_add if x not in existing_perms.permissions_any}

        if not existing_perms.check_permissions(msg.author, msg.channel):
            raise UserPermissionError

        new_perms = CommandPermissions.from_existing(existing_perms,
                                                     permissions_any=existing_perms.permissions_any | perms_to_add)

        if not new_perms.check_permissions(msg.author, msg.channel):
            raise UserPermissionError

        command_to_edit.perms.update(new_perms)
        self.permission_overrides[command_to_edit.name.lower()]["permissions_any"] = list(new_perms.permissions_any)
        self.config_manager.save_config()

        await respond(msg, f"**ANALYSIS: Command {command_to_edit.name} edited successfully. Added \"any of\" "
                           f"permissions:** `{', '.join(successfully_added_perms)}`")

    @Command("RemoveCommandPermissionAnyOf", "RmCmdPermAny",
             doc="Removes an \"any of\" permission from a given command. If all required permissions are removed "
                 "from a command, and it has no \"any of\" permissions, it can be used by anyone.",
             syntax="(command) (permissions_to_remove)",
             perms="manage_guild",
             category="command_permissions")
    async def _rm_command_permission_any(self, msg: discord.Message):
        command_to_edit, perms_to_remove = self._validate_arguments_permissions(msg.clean_content)
        existing_perms = command_to_edit.perms

        if not existing_perms.check_permissions(msg.author, msg.channel):
            raise UserPermissionError

        new_perms = CommandPermissions.from_existing(existing_perms,
                                                     permissions_any=existing_perms.permissions_any - perms_to_remove)
        difference_set = existing_perms.permissions_any - new_perms.permissions_any

        command_to_edit.perms.update(new_perms)
        self.permission_overrides[command_to_edit.name.lower()]["permissions_any"] = list(new_perms.permissions_any)
        self.config_manager.save_config()

        await respond(msg, f"**ANALYSIS: Command {command_to_edit.name} edited successfully. Removed \"any of\" "
                           f"permissions:** `{', '.join(x.upper() for x in difference_set)}`")

    @Command("AddCommandUserOverride", "AddCmdUser",
             doc="Adds a user override(s) to a given command. A user with an override can use a command as if they "
                 "have all required and optional permissions.",
             syntax="(command) (users_to_add)",
             perms="manage_guild",
             category="command_permissions")
    async def _add_command_user_override(self, msg: discord.Message):
        command_to_edit, users_to_add = self._validate_arguments_users(msg.clean_content)
        existing_perms = command_to_edit.perms
        ids_to_add = {x.id for x in users_to_add}
        successfully_added_users = {x.name for x in users_to_add if x.id not in existing_perms.user_overrides}

        if not existing_perms.check_permissions(msg.author, msg.channel):
            raise UserPermissionError

        new_perms = CommandPermissions.from_existing(existing_perms,
                                                     user_overrides=existing_perms.user_overrides | ids_to_add)

        command_to_edit.perms.update(new_perms)
        self.permission_overrides[command_to_edit.name.lower()]["user_overrides"] = list(new_perms.user_overrides)
        self.config_manager.save_config()

        await respond(msg, f"**ANALYSIS: Command {command_to_edit.name} edited successfully. Added user overrides:** "
                           f"`{', '.join(successfully_added_users)}`")

    @Command("RemoveCommandUserOverride", "RmCmdUser",
             doc="Removes a user override(s) from a given command.",
             syntax="(command) (users_to_remove)",
             perms="manage_guild",
             category="command_permissions")
    async def _rm_command_user_override(self, msg: discord.Message):
        command_to_edit, users_to_remove = self._validate_arguments_users(msg.clean_content)
        existing_perms = command_to_edit.perms
        ids_to_remove = {x.id for x in users_to_remove}
        successfully_removed_users = {x.name for x in users_to_remove if x in existing_perms.user_overrides}

        if not existing_perms.check_permissions(msg.author, msg.channel):
            raise UserPermissionError

        new_perms = CommandPermissions.from_existing(existing_perms,
                                                     user_overrides=existing_perms.user_overrides - ids_to_remove)

        if not new_perms.check_permissions(msg.author, msg.channel):
            raise UserPermissionError

        command_to_edit.perms.update(new_perms)
        self.permission_overrides[command_to_edit.name.lower()]["user_overrides"] = list(new_perms.user_overrides)
        self.config_manager.save_config()

        await respond(msg, f"**ANALYSIS: Command {command_to_edit.name} edited successfully. Removed user "
                           f"overrides:** `{', '.join(successfully_removed_users)}`")

    @Command("AddCommandRoleOverride", "AddCmdRole",
             doc="Adds a role override(s) to a given command. Users with any of the roles in the override list can "
                 "use a command as if they have all required and optional permissions.",
             syntax="(command) (roles_to_add)",
             perms="manage_guild",
             category="command_permissions")
    async def _add_command_role_override(self, msg: discord.Message):
        command_to_edit, roles_to_add = self._validate_arguments_roles(msg.clean_content)
        existing_perms = command_to_edit.perms
        ids_to_add = {x.id for x in roles_to_add}
        successfully_added_roles = {x.name for x in roles_to_add if x.id not in existing_perms.role_overrides}

        if not existing_perms.check_permissions(msg.author, msg.channel):
            raise UserPermissionError

        new_perms = CommandPermissions.from_existing(existing_perms,
                                                     role_overrides=existing_perms.role_overrides | ids_to_add)

        command_to_edit.perms.update(new_perms)
        self.permission_overrides[command_to_edit.name.lower()]["role_overrides"] = list(new_perms.role_overrides)
        self.config_manager.save_config()

        await respond(msg, f"**ANALYSIS: Command {command_to_edit.name} edited successfully. Added role overrides:** "
                           f"`{', '.join(successfully_added_roles)}`")

    @Command("RemoveCommandRoleOverride", "RmCmdRole",
             doc="Removes a role override(s) from a given command.",
             syntax="(command) (roles_to_remove)",
             perms="manage_guild",
             category="command_permissions")
    async def _rm_command_role_override(self, msg: discord.Message):
        command_to_edit, roles_to_remove = self._validate_arguments_roles(msg.clean_content)
        existing_perms = command_to_edit.perms
        ids_to_remove = {x.id for x in roles_to_remove}
        successfully_removed_roles = {x.name for x in roles_to_remove if x in existing_perms.role_overrides}

        if not existing_perms.check_permissions(msg.author, msg.channel):
            raise UserPermissionError

        new_perms = CommandPermissions.from_existing(existing_perms,
                                                     role_overrides=existing_perms.role_overrides - ids_to_remove)

        if not new_perms.check_permissions(msg.author, msg.channel):
            raise UserPermissionError

        command_to_edit.perms.update(new_perms)
        self.permission_overrides[command_to_edit.name.lower()]["role_overrides"] = list(new_perms.role_overrides)
        self.config_manager.save_config()

        await respond(msg, f"**ANALYSIS: Command {command_to_edit.name} edited successfully. Removed role "
                           f"overrides:** `{', '.join(successfully_removed_roles)}`")

    @Command("AddCommandOptionalPermission", "AddCmdPermOpt",
             doc="Adds a permission to a given command's specified optional permission set. Optional permission sets "
                 "are used to give commands additional functions for users who have all of the permissions in the "
                 "set - such as bypassing limits or modifying other members. Most commands don't have these - refer "
                 "to the command of interest's documentation for more information.",
             syntax="(command) (feature_set) (permissions_to_add)",
             perms="manage_guild",
             category="command_permissions")
    async def _add_command_optional_permission(self, msg: discord.Message):
        command_to_edit, feature_set, perms_to_add = self._validate_arguments_optional_permissions(msg.clean_content)
        existing_perms = command_to_edit.perms
        successfully_added_perms = {x.upper() for x in perms_to_add
                                    if x not in existing_perms.optional_permissions[feature_set]}

        if not existing_perms.check_optional_permissions(feature_set, msg.author, msg.channel):
            raise UserPermissionError

        new_optional_perms = existing_perms.optional_permissions.copy()
        new_optional_perms[feature_set] = new_optional_perms[feature_set].copy() | perms_to_add
        new_perms = CommandPermissions.from_existing(existing_perms, optional_permissions=new_optional_perms)

        if not new_perms.check_optional_permissions(feature_set, msg.author, msg.channel):
            raise UserPermissionError

        command_to_edit.perms.update(new_perms)
        self.permission_overrides[command_to_edit.name.lower()]["optional_permissions"] = \
            {k: list(v) for k, v in new_perms.optional_permissions.items()}
        self.config_manager.save_config()

        await respond(msg, f"**ANALYSIS: Command {command_to_edit.name} edited successfully. Added permissions to "
                           f"optional feature set {feature_set}:** `{', '.join(successfully_added_perms)}`")

    @Command("RemoveCommandOptionalPermission", "RmCmdPermOpt",
             doc="Removes a permission from the given command's specified optional permission set. If all "
                 "permissions are removed from a feature set, all users have access to the optional features.",
             syntax="(command) (feature_set) (permissions_to_add)",
             perms="manage_guild",
             category="command_permissions")
    async def _rm_command_optional_permission(self, msg: discord.Message):
        command_to_edit, feature_set, perms_to_remove = \
            self._validate_arguments_optional_permissions(msg.clean_content)
        existing_perms = command_to_edit.perms
        successfully_removed_perms = {x.upper() for x in perms_to_remove if x in
                                      existing_perms.optional_permissions[feature_set]}

        if not existing_perms.check_optional_permissions(feature_set, msg.author, msg.channel):
            raise UserPermissionError

        new_optional_perms = existing_perms.optional_permissions.copy()
        new_optional_perms[feature_set] = new_optional_perms[feature_set].copy() - perms_to_remove
        new_perms = CommandPermissions.from_existing(existing_perms, optional_permissions=new_optional_perms)

        command_to_edit.perms.update(new_perms)
        self.permission_overrides[command_to_edit.name.lower()]["optional_permissions"] = \
            {k: list(v) for k, v in new_perms.optional_permissions.items()}
        self.config_manager.save_config()

        await respond(msg, f"**ANALYSIS: Command {command_to_edit.name} edited successfully. Removed permissions "
                           f"from optional feature set {feature_set}:** `{', '.join(successfully_removed_perms)}`")

    @Command("ListCommandPermissions", "ListCmdPerms",
             doc="Lists the permission configuration for a given command.",
             syntax="(command)",
             perms="manage_guild",
             category="command_permissions")
    async def _list_command_permissions(self, msg: discord.Message):
        args = shlex.split(msg.clean_content.lower())[1:]
        try:
            command_to_read = self._validate_arguments_command(args.pop(0))
        except IndexError:
            raise CommandSyntaxError

        perms = command_to_read.perms

        if perms.permissions_all:
            permissions_all_printable = ', '.join(x.upper() for x in perms.permissions_all)
        else:
            permissions_all_printable = "None"
        permissions_all_printable = f"Required permissions: `{permissions_all_printable}`\n"
        if perms.permissions_any:
            permissions_any_printable = f"Any permission of these required: " \
                                        f"`{', '.join(x.upper() for x in perms.permissions_any)}`\n"
        else:
            permissions_any_printable = ""
        if perms.user_overrides:
            users_printable = f"Users with permission override: " \
                              f"`{', '.join(self.guild.get_member(x).name for x in perms.user_overrides)}`\n"
        else:
            users_printable = ""
        if perms.role_overrides:
            roles_printable = f"Roles with permission overrides: " \
                              f"`{', '.join(self.guild.get_role(x).name for x in perms.role_overrides)}`\n"
        else:
            roles_printable = ""

        if perms.optional_permissions:
            optionals_printable = "Optional feature sets:\n"
            optionals_printable += "\n".join(f"- {k}: {', '.join(x.upper() for x in v)}"
                                             for k, v in perms.optional_permissions.items())
            optionals_printable += "\n"
        else:
            optionals_printable = ""

        response = f"```Permissions for command {command_to_read.name}:\n" \
                   f"{permissions_all_printable}{permissions_any_printable}{users_printable}{roles_printable}" \
                   f"{optionals_printable}```"

        await respond(msg, response)

    # Utility functions

    def _validate_arguments_command(self, command: str) -> Command:
        try:
            return self.commands[command]
        except KeyError:
            raise CommandSyntaxError(f"Command {command} does not exist.")

    def _validate_arguments_permissions(self, args: str) -> tuple[Command, set[str]]:
        args = shlex.split(args.lower())[1:]
        try:
            command_to_edit = self._validate_arguments_command(args.pop(0))
        except IndexError:
            raise CommandSyntaxError

        if not all(hasattr(discord.Permissions, x) for x in args):
            raise CommandSyntaxError("One or more permissions do not exist.")

        return command_to_edit, set(args)

    def _validate_arguments_optional_permissions(self, args: str) -> tuple[Command, str, set[str]]:
        args = shlex.split(args.lower())[1:]
        try:
            command_to_edit = self._validate_arguments_command(args.pop(0))
        except IndexError:
            raise CommandSyntaxError

        try:
            feature_set_to_edit = args.pop(0)
        except IndexError:
            raise CommandSyntaxError
        try:
            feature_set_to_edit = command_to_edit.perms.optional_permissions[feature_set_to_edit]
        except KeyError:
            raise CommandSyntaxError(f"Command {command_to_edit.name} has no such optional feature set "
                                     f"{feature_set_to_edit}.")

        if not all(hasattr(discord.Permissions, x) for x in args):
            raise CommandSyntaxError("One or more permissions do not exist.")

        return command_to_edit, feature_set_to_edit, set(args)

    def _validate_arguments_users(self, args: str) -> tuple[Command, set[discord.Member]]:
        args = shlex.split(args.lower())[1:]
        try:
            command_to_edit = self._validate_arguments_command(args.pop(0))
        except IndexError:
            raise CommandSyntaxError

        users = set()
        for search_str in args:
            user = find_user(self.guild, search_str)
            if user is None:
                raise CommandSyntaxError(f"User {search_str} could not be found.")
            users.add(user)

        return command_to_edit, users

    def _validate_arguments_roles(self, args: str) -> tuple[Command, set[discord.Role]]:
        args = shlex.split(args.lower())[1:]
        try:
            command_to_edit = self._validate_arguments_command(args.pop(0))
        except IndexError:
            raise CommandSyntaxError

        roles = set()
        for search_str in args:
            role = find_role(self.guild, search_str)
            if role is None:
                raise CommandSyntaxError(f"Role {search_str} could not be found.")
            roles.add(role)

        return command_to_edit, roles



from __future__ import annotations
from red_star.command_dispatcher import Command
from red_star.plugin_manager import BasePlugin
from red_star.rs_errors import CommandSyntaxError
from red_star.rs_utils import respond, decode_json, split_message, verify_embed
from urllib.parse import urlparse
from urllib.request import urlopen
from urllib.error import URLError
import mimetypes
import re
import json
import discord
from io import BytesIO


def verify_document(doc: list):
    """
    A helper function to verify entire documents.
    Verifies that messages are of correct length, embeds are valid and file links are non-mangled.
    :param doc:
    :return:
    """
    result = []
    for i, msg in enumerate(doc):
        if isinstance(msg, str):
            if len(msg) > 2000:
                raise ValueError(f"Message {i+1} is too long. (Limit 2000)")
            result.append(msg)
        else:
            try:
                _msg = dict()
                if 'content' in msg:
                    if len(str(msg['content'])) > 2000:
                        raise ValueError(f"Message {i+1} is too long. (Limit 2000)")
                    _msg['content'] = str(msg['content'])
                if 'embed' in msg:
                    try:
                        _msg['embed'] = verify_embed(doc[i]['embed'])
                    except ValueError as e:
                        raise ValueError(f"Message {i+1} invalid embed: {e}")

                if 'file' in msg:
                    url = urlparse(msg['file'])
                    if not (url.scheme and url.netloc and url.path):
                        raise ValueError(f"Message {i+1} invalid attach url.")
                    _msg['file'] = msg['file']
                result.append(_msg)
            except TypeError:
                raise ValueError(f"Message {i+1} not an object or string.")
    return result


def verify_links(doc: list, max_size: int):
    """
    A secondary helper functions for the purpose of checking all the links for both existing and being within limit.
    :param doc: document, a list of messages or dicts
    :param max_size: maximum size of a file in octets
    :return:
    """
    for i, msg in enumerate(doc):
        try:
            _file = urlopen(msg['file'])
            if int(_file.info()['Content-Length']) > max_size:
                raise ValueError(f"Message {i+1} attached file too big.")
        except (TypeError, KeyError):
            continue
        except ValueError as e:
            raise ValueError(e)
        except Exception as e:
            raise ValueError(f"Message {i+1} dead attachment link: {e}.")


class ChannelPrint(BasePlugin):
    name = "channel_print"
    version = "1.0"
    author = "GTG3000"
    description = "A plugin that allows printing of prepared multi-message data."

    default_global_config = {
        "max_filesize": 1024 * 1024 * 8  # max 8 mb
    }
    log_events = {"print_event"}

    async def activate(self):
        self._port_old_storage()
        self.storage.setdefault("documents", {})

    def _port_old_storage(self):
        old_storage_path = self.config_manager.config_path / "walls.json"
        if old_storage_path.exists():
            with old_storage_path.open(encoding="utf-8") as fp:
                old_storage = json.load(fp)
            for guild_id, walls in old_storage.items():
                try:
                    new_storage = self.config_manager.storage_files[guild_id][self.name]
                except KeyError:
                    self.logger.warn(f"Server with ID {guild_id} not found! Is the bot still in this server?\n"
                                     f"Skipping conversion of this server's channel print document storage...")
                    continue
                new_storage.contents["documents"] = walls
                new_storage.save()
                new_storage.load()
            old_storage_path = old_storage_path.replace(old_storage_path.with_suffix(".json.old"))
            self.logger.info(f"Old channel print document storage converted to new format. "
                             f"Old data now located at {old_storage_path} - you may delete this file.")

    def storage_save_args(self):
        return {'indent': 2, 'ensure_ascii': False}

    @Command("Print", "PrintForce",
             doc="Prints out the specified document from the storage, allowing to dump large amounts of information "
                 "into a channel, for example for purposes of a rules channel.\n"
                 "Document can be specified from the saved ones, or uploaded with the command.\n"
                 "Use \"PrintForce\" alias to force printing despite broken attachment links.",
             syntax="(document)",
             perms={"manage_messages"},
             category="channel_print",
             run_anywhere=True,
             delete_call=True)
    async def _print(self, msg: discord.Message):
        cmd, *args = msg.clean_content.split(None, 1)

        if msg.attachments:
            _file = BytesIO()

            await msg.attachments[0].save(_file)
            try:
                wall = decode_json(_file.getvalue())
            except json.JSONDecodeError as e:
                raise CommandSyntaxError(f"Not a valid JSON file: {e}")
            except ValueError as e:
                self.logger.exception("Could not decode uploaded document file!", exc_info=True)
                raise CommandSyntaxError(e)
        else:
            try:
                wall = args[0]
            except IndexError:
                raise CommandSyntaxError

            if wall not in self.storage["documents"]:
                raise CommandSyntaxError("No such document.")
            wall = self.storage["documents"][wall]

        try:
            wall = verify_document(wall)
            if not cmd.lower().endswith("force"):
                verify_links(wall, self.global_plugin_config['max_filesize'])
        except ValueError as e:
            raise CommandSyntaxError(e)

        for post in wall:
            if isinstance(post, str):
                await respond(msg, post)
            else:
                _file = None
                if 'file' in post:
                    try:
                        _file = urlopen(post['file'])
                        if int(_file.info()['Content-Length']) > self.global_plugin_config['max_filesize']:
                            raise ValueError("File too big.")
                        ext = mimetypes.guess_extension(_file.info()['Content-Type'])
                        _file = discord.File(_file, filename="wallfile" + ext)
                    except (URLError, TypeError, ValueError) as e:
                        self.logger.info(f"Attachment file error in {msg.guild}:\n{e}")
                        await self.plugin_manager.hook_event("on_log_event",
                                                             f"**WARNING: Error occurred during printout:**\n{e}",
                                                             log_type="print_event")
                        _file = None  # Just fail silently if the request doesn't work out
                if _file or 'embed' in post or 'content' in post:
                    await respond(msg,
                                  post.get('content', None),
                                  embed=post['embed'] if 'embed' in post else None,
                                  file=_file)

    @Command("DeletePrint", "PrintDelete",
             doc="Deletes the specified document.",
             syntax="(document)",
             perms={"manage_messages"},
             category="channel_print")
    async def _delete_print(self, msg: discord.Message):
        try:
            name = msg.clean_content.split(None, 2)[1].lower()
            del self.storage["documents"][name]
            self.storage_file.save()
            await respond(msg, "**AFFIRMATIVE. Document deleted.**")
        except IndexError:
            raise CommandSyntaxError("Document name required.")
        except KeyError:
            raise CommandSyntaxError("No document found.")

    @Command("ListPrint", "PrintList",
             doc="Lists all available documents.",
             perms={"manage_messages"},
             category="channel_print")
    async def _listprint(self, msg: discord.Message):
        walls = "\n".join(self.storage["documents"].keys())
        final_msg = f"**ANALYSIS: Following documents are available:**```\n{walls}```"
        for split_msg in split_message(final_msg):
            await respond(msg, split_msg)

    @Command("DumpPrint", "PrintDump",
             doc="Uploads the specified document in a json file format.",
             syntax="(document)",
             perms={"manage_messages"},
             category="channel_print")
    async def _dump_print(self, msg: discord.Message):
        try:
            name = msg.clean_content.split(None, 2)[1].lower()
            dump_data = bytes(json.dumps(self.storage["documents"][name], indent=2, ensure_ascii=False),
                              encoding="utf8")
            await respond(msg, "**AFFIRMATIVE. Uploading file.**",
                          file=discord.File(BytesIO(dump_data), filename=name+'.json'))
        except IndexError:
            raise CommandSyntaxError("Document name required.")
        except KeyError:
            raise CommandSyntaxError("No document found.")

    @Command("UploadPrint", "PrintUpload",
             doc="Allows you to upload a document, in a JSON file format or a JSON code block.",
             syntax="(document_id) (code block or attached file)",
             perms={"manage_messages"},
             category="channel_print")
    async def _upload_print(self, msg: discord.Message):
        try:
            name = msg.clean_content.split(None, 2)[1].lower()
        except IndexError:
            raise CommandSyntaxError("Document name required.")

        if msg.attachments:
            _file = BytesIO()

            await msg.attachments[0].save(_file)
            try:
                data = decode_json(_file.getvalue())
            except ValueError as e:
                self.logger.exception("Could not decode uploaded document file!", exc_info=True)
                raise CommandSyntaxError(e)
            except Exception as e:
                raise CommandSyntaxError(f"Not a valid JSON file: {e}")
        else:
            try:
                data = re.match(r"```.*?(?P<json>\[.+]).*?```", msg.clean_content.split(None, 2)[2], re.DOTALL)['json']
                data = json.loads(data)
            except IndexError:
                raise CommandSyntaxError("JSON code block required.")
            except TypeError:
                raise CommandSyntaxError("Invalid JSON template. The root must be a list.")
            except Exception as e:
                raise CommandSyntaxError(f"Not a valid JSON file: {e}")

        try:
            data = verify_document(data)
        except ValueError as e:
            raise CommandSyntaxError(e)

        self.storage["documents"][name] = data
        self.storage_file.save()

        await respond(msg, f"**AFFIRMATIVE. Document {name} available for printout.**")

    @Command("PrintReload",
             doc="Reloads all documents from list. You probably shouldn't be using this too often.",
             bot_maintainers_only=True,
             category="channel_print")
    async def _print_reload(self, msg: discord.Message):
        self.storage.load()
        await respond(msg, "**AFFIRMATIVE. Printout documents reloaded.**")



from __future__ import annotations
import datetime
import json
import discord.utils
from discord.ext import tasks
from random import choice
from red_star.plugin_manager import BasePlugin
from red_star.rs_errors import CommandSyntaxError, ChannelNotFoundError, DataCarrier
from red_star.rs_utils import respond, RSArgumentParser
from red_star.command_dispatcher import Command

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    import discord


class MOTD(BasePlugin):
    name = "motd"
    version = "1.4"
    author = "medeor413"
    description = "A plugin for flexibly displaying messages at date change."
    default_config = {
        "default": {
            "motd_file": "motds.json"
        }
    }
    channel_types = {"motd"}

    valid_months = ["jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec"]
    month_names_to_abbr = {"january": "jan", "february": "feb", "march": "mar", "april": "apr", "may": "may",
                           "june": "jun", "july": "jul", "august": "aug", "september": "sep", "october": "oct",
                           "november": "nov", "december": "dec"}
    valid_weekdays = {"mon", "tue", "wed", "thu", "fri", "sat", "sun"}
    valid_days = {str(i) for i in range(1, 32)}
    valid_month_weeks = {"week-1", "week-2", "week-3", "week-4", "week-5", "week-6", "last-week"}
    valid_dates = set(valid_months) | valid_month_weeks | valid_weekdays | valid_days

    async def activate(self):
        self._port_old_storage()
        self._display_motd.start()

    def _port_old_storage(self):
        motds_folder = self.client.storage_dir / "motds"
        if "motd_file" not in self.config:
            return
        old_motd_file = motds_folder / self.config["motd_file"]
        if old_motd_file.exists():
            with old_motd_file.open() as fp:
                old_motds = json.load(fp)
            self.storage["motds"] = old_motds
            del self.config["motd_file"]
            self.config_manager.save_config()
            self.storage_file.save()
            self.storage_file.load()
            self.logger.info("Ported MotD file to new storage system.")

    @tasks.loop(time=discord.utils.utcnow().replace(hour=0, minute=0, second=0).timetz())
    async def _display_motd(self, date: datetime.date):
        try:
            chan = self.channel_manager.get_channel("motd")
        except ChannelNotFoundError:
            return
        try:
            lines = self._get_motds(date)
        except DataCarrier as dc:
            lines = dc.data
        try:
            line = choice(lines)
            line = date.strftime(line)
        except IndexError:
            return
        except ValueError:
            line = choice(lines)
        await chan.send(line)

    def _get_motds(self, date: datetime.date | None, options: dict = None, valid: set = None):
        if options is None:
            options = self.storage["motds"]
        if not valid:
            valid = {date.strftime("%b").lower(), date.strftime("%a").lower(),
                     str(date.day), "week-" + str(week_of_month(date))}
            if is_last_week_of_month(date):
                valid.add("last-week")
        results = []
        if options.get("holiday", False):
            raise DataCarrier(options["options"])
        for k, v in options.items():
            if k.lower() in valid:
                results.extend(self._get_motds(date, options=v, valid=valid))
        if "options" in options:
            results.extend(options["options"])
        return results

    @Command("AddMotD",
             doc="Adds a MotD message. Each element of the path must match the date for an MotD to display.\n"
                 "Use --holiday to declare a holiday; when a holiday MotD list is checked, it will only use MotDs "
                 "from that specific list.",
             perms={"manage_guild"},
             category="bot_management",
             syntax="[-h/--holiday] (/path/of/date) (message)")
    async def _add_motd(self, msg: discord.Message):
        holiday = False
        try:
            path, new_motd = msg.clean_content.split(None, 2)[1:]
            self.logger.debug(f"New MOTD for {path}: {new_motd}")
            if path in ("-h", "--holiday"):
                holiday = True
                path, new_motd = new_motd.split(None, 1)
        except ValueError:
            raise CommandSyntaxError("Not enough arguments.")

        path = [self.month_names_to_abbr.get(x, x) for x in path.lower().split("/") if x]
        if not all(x in self.valid_dates for x in path):
            invalid = ", ".join(x for x in path if x not in self.valid_dates)
            raise CommandSyntaxError(f"The following path identifiers are invalid: `{invalid}`")
        motd_path = self.storage["motds"]
        for k in path:
            motd_path = motd_path.setdefault(k, {})
        motd_path.setdefault("options", []).append(new_motd)
        if holiday:
            motd_path["holiday"] = True

        self.storage_file.save()
        await respond(msg, f"**ANALYSIS: MotD for {'/'.join(path)} added successfully"
                           f"{' and set as holiday' if holiday else ''}.**")

    @Command("TestMotDs",
             doc="Used for testing MOTD lines. If a date is not provided, today's date will be used. Month may be "
                 "specified by full English name, three-letter abbreviation, or number. If --dt is used, "
                 "don't use any other flags.",
             perms={"manage_guild"},
             category="debug",
             syntax="[-d/--day number] [-wd/--weekday weekday] [-mw/--monthweek week-x] [-m/--month month] "
                    "OR [-dt/--dt ISO-format date]")
    async def _test_motd(self, msg: discord.Message):
        today = discord.utils.utcnow()
        parser = RSArgumentParser()
        parser.add_argument("-d",  "--day", default=str(today.day))
        parser.add_argument("-wd", "--weekday", default=today.strftime("%a").lower())
        parser.add_argument("-mw", "--monthweek", default="week-" + str(week_of_month(today)))
        parser.add_argument("-m",  "--month", default=today.strftime("%b").lower(), type=self.coerce_month_value)
        parser.add_argument("-dt", "--date", default=None, type=datetime.datetime.fromisoformat)
        args = parser.parse_args(msg.clean_content.lower().split()[1:])

        try:
            if args.date:
                lines = self._get_motds(args.date)
            else:
                if args.month not in self.valid_months \
                        or args.day not in self.valid_days \
                        or args.weekday not in self.valid_weekdays \
                        or args.monthweek not in self.valid_month_weeks:
                    raise CommandSyntaxError("One of the arguments is not valid.")
                lines = self._get_motds(None, valid={args.month, args.day, args.weekday, args.monthweek})
        except DataCarrier as dc:
            lines = dc.data
        lines = "\n".join(lines)
        if not lines:
            lines = "None."
        await respond(msg, f"```\n{lines}```")

    def coerce_month_value(self, month):
        if month in self.valid_months:
            return month
        elif month in self.month_names_to_abbr:
            return self.month_names_to_abbr[month]
        elif month.isdigit() and 1 <= int(month) <= 12:
            return self.valid_months[int(month) - 1]
        raise ValueError(f"{month} is not a valid month.")


def week_of_month(date: datetime.date):
    first_of_month = date.replace(day=1).weekday()
    if first_of_month == 6:
        first_of_month -= 7
    return (date.day + first_of_month) // 7 + 1
    

def is_last_week_of_month(date: datetime.date):
    last_of_month = date.replace(month=(date.month % 12 + 1), day=1) - datetime.timedelta(days=1)
    return week_of_month(date) == week_of_month(last_of_month)



from __future__ import annotations

import json
import logging

from red_star.rs_errors import ChannelNotFoundError

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    import discord
    from red_star.client import RedStar


class ChannelManager:
    name = "channel_manager"
    channel_types = {"commands"}
    channel_categories = {"no_read"}

    def __init__(self, client: RedStar, guild: discord.Guild):
        self.client = client
        self.guild = guild
        self.logger = logging.getLogger(f"red_star.{self.name}.{guild.id}")
        self.config_manager = client.config_manager
        self.default_config = {
            "channels": {},
            "categories": {}
        }
        self.storage_file = self.config_manager.get_plugin_storage(self)

        self._port_old_storage()

        if not self.conf:
            self.conf = {
                "channels": {i: None for i in self.channel_types},
                "categories": {i: [] for i in self.channel_categories}
            }

        if not set(self.conf["channels"].keys()) >= self.channel_types:
            defaults_added_dict = {x: None for x in self.channel_types}.update(self.conf["channels"])
            self.conf["channels"] = defaults_added_dict
        if not set(self.conf["categories"].keys()) >= self.channel_categories:
            defaults_added_dict = {x: [] for x in self.channel_categories}.update(self.conf["categories"])
            self.conf["categories"] = defaults_added_dict
        self.storage_file.save()

    @property
    def conf(self):
        return self.storage_file.contents

    @conf.setter
    def conf(self, value):
        self.storage_file.contents = value

    def _port_old_storage(self):
        old_storage_path = self.config_manager.config_path / "channel_manager.json"
        if old_storage_path.exists():
            with old_storage_path.open(encoding="utf-8") as fp:
                old_storage = json.load(fp)
            for guild_id, channel_data in old_storage.items():
                try:
                    new_storage = self.config_manager.storage_files[guild_id][self.name]
                except KeyError:
                    self.logger.warning(f"Server with ID {guild_id} not found! Is the bot still in this server?\n"
                                        f"Skipping conversion of this server's channel manager storage...")
                    continue
                new_storage.contents = channel_data
                new_storage.save()
                new_storage.load()
            old_storage_path = old_storage_path.replace(old_storage_path.with_suffix(".json.old"))
            self.logger.info(f"Old channel manager storage converted to new format. "
                             f"Old data now located at {old_storage_path} - you may delete this file.")

    def storage_save_args(self):
        return {}

    def storage_load_args(self):
        return {}

    def get_channel(self, channel_type: str):
        channel_type = channel_type.lower()
        chan = self.guild.get_channel(self.conf["channels"][channel_type])
        if not chan:
            raise ChannelNotFoundError(channel_type)
        return chan

    def set_channel(self, channel_type: str, channel: discord.abc.GuildChannel):
        channel_type = channel_type.lower()
        if channel:
            self.conf["channels"][channel_type] = channel.id
        else:
            self.conf["channels"][channel_type] = None
        self.storage_file.save()

    def get_category(self, category: str):
        return self.conf["categories"].get(category.lower())

    def channel_in_category(self, category: str, channel: discord.abc.GuildChannel):
        return channel in self.conf["categories"].get(category.lower(), {})

    def add_channel_to_category(self, category: str, channel: discord.abc.GuildChannel):
        category = self.conf["categories"].setdefault(category.lower(), [])
        if channel.id not in category:
            category.append(channel.id)
            self.storage_file.save()
            return True
        else:
            return False

    def remove_channel_from_category(self, category: str, channel: discord.abc.GuildChannel):
        category = category.lower()
        if self.channel_in_category(category, channel):
            self.conf["categories"][category].remove(channel.id)
            self.storage_file.save()
            return True
        else:
            return False



from red_star.plugin_manager import BasePlugin
from red_star.command_dispatcher import Command
from red_star.rs_utils import respond, RSArgumentParser, split_message, find_role, group_items
from red_star.rs_errors import CommandSyntaxError
import asyncio
import discord
import shlex
import json


class MESSAGE_TYPE:
    NORMAL = 0
    CHOICE = 1


class RoleRequest(BasePlugin):
    name = "role_request"
    version = "1.4"
    author = "GTG3000"

    default_plugin_config = {
        "roles": [],
        "default_roles": []
    }

    reacts = None
    passwords = None

    async def activate(self):
        self._port_old_storage()
        self.reacts = self.storage.setdefault("role_request_reaction_messages", {})
        self.passwords = self.storage.setdefault("role_password", {})
        self._port_old_reacts()

    def _port_old_reacts(self):
        for messageId, config in self.reacts.items():
            if type(config) == list:
                self.reacts[messageId] = {
                    "reacts": config,
                    "type": MESSAGE_TYPE.NORMAL
                }
            elif "required" not in config:
                config["required"] = []

    def _port_old_storage(self):
        old_storage_path = self.config_manager.config_path / "role_request_reaction_messages.json"
        if old_storage_path.exists():
            with old_storage_path.open(encoding="utf-8") as fp:
                old_storage = json.load(fp)
            for msg_id, message_data in old_storage.items():
                # We have no way to check which guild a message belongs to, so we have no choice but to put all
                # message IDs in all guilds.
                for new_storage in self.config_manager.storage_files.values():
                    plugin_storage = new_storage[self.name]
                    plugin_storage.contents.setdefault("role_request_reaction_messages", {})[msg_id] = message_data
                    plugin_storage.save()
                    plugin_storage.load()
            old_storage_path = old_storage_path.replace(old_storage_path.with_suffix(".json.old"))
            self.logger.info(f"Old role request reaction message storage converted to new format. "
                             f"Old data now located at {old_storage_path} - you may delete this file.")

    async def on_member_join(self, member: discord.Member):
        """
        Handles the call of on_member_join to apply the default roles, if any.
        """
        roles = self.config['default_roles']
        if roles:
            roles = map(lambda rid: find_role(member.guild, str(rid)), roles)
            await member.add_roles(*roles, reason="Adding default roles.")

    @Command("DefaultRole",
             doc="-a/--add   : Adds specified roles to the list of default roles.\n"
                 "-r/--remove: Removes specified roles from the list.\n"
                 "Calling it without any arguments prints the list.",
             syntax="[-a/--add (role mentions/ids/names)] [-r/--remove (role mentions/ids/names)]",
             perms={"manage_roles"},
             category="role_request")
    async def _manage_default(self, msg: discord.Message):
        parser = RSArgumentParser()
        parser.add_argument("command")
        parser.add_argument("-a", "--add", default=[], nargs='+')
        parser.add_argument("-r", "--remove", default=[], nargs='+')

        args = parser.parse_args(shlex.split(msg.content))

        d_r_list = self.config["default_roles"]

        if not (args['add'] or args['remove']):
            role_str = "\n".join(x.name for x in msg.guild.roles if x.id in d_r_list)
            for split_msg in split_message(f"**ANALYSIS: Currently approved default roles:```\n{role_str}```"):
                await respond(msg, split_msg)
        else:
            args['add'] = [r for r in [find_role(msg.guild, r) for r in args['add']] if r]
            args['remove'] = [r for r in [find_role(msg.guild, r) for r in args['remove']] if r]

            # for nice output
            added_roles = []
            removed_roles = []

            for role in args['add']:
                if role.id not in d_r_list:
                    added_roles.append(role.name)
                    d_r_list.append(role.id)
            for role in args['remove']:
                if role.id in d_r_list:
                    removed_roles.append(role.name)
                    d_r_list.remove(role.id)

            if added_roles or removed_roles:
                output_str = "**AFFIRMATIVE. ANALYSIS:**\n```diff\n"
                if added_roles:
                    output_str += "Added roles:\n+ " + "\n+ ".join(added_roles) + "\n"
                if removed_roles:
                    output_str += "Removed roles:\n- " + "\n- ".join(removed_roles) + "\n"
                output_str += "```"
                await respond(msg, output_str)
            else:
                raise CommandSyntaxError

    @Command("OfferRoles",
             doc="Creates a post with reactions that can be used to gain or remove offered roles.\n"
                 "Use as follows: OfferRoles 🎈 balloon-role 📌 \"pushpin role\"\n"
                 "Optional flags:\n"
                 "-x/--exclusive: only one reaction is allowed to be claimed.\n"
                 "-r/--require  : member must have all of the required roles to claim anything.\n\n"
                 "example: OfferRoles -x -r Speaking -r Special ✅ @extra_special_role",
             syntax="[-x/--exclusive] [-r/--require (role)] (reaction) (role)...",
             perms={"manage_roles"},
             category="role_request",
             run_anywhere=True)
    async def _offer_roles(self, msg: discord.Message):
        parser = RSArgumentParser()
        parser.add_argument("command")
        parser.add_argument("-x", "--exclusive", action="store_true")
        parser.add_argument("-r", "--require", action="append", default=[])
        parser.add_argument("pairs", nargs="+")

        args = parser.parse_args(shlex.split(msg.content))

        if not args["pairs"]:
            return

        print(args)

        found = list(filter(lambda x: x[1],  # throw out the entries where role wasn't found
                            ((emote, find_role(msg.guild, roleQuery))
                             for emote, roleQuery in zip(args["pairs"][::2], args["pairs"][1::2]))))

        requisite = []
        for roleQuery in args["require"]:
            role = find_role(msg.guild, roleQuery)
            if not role:
                raise CommandSyntaxError(F"Could not find required role {roleQuery}")
            requisite.append(role.id)

        if found:
            message = await respond(msg, "**Following roles available through reacting to this message:**\n" +
                                    "\n".join(f"> {react}: {r.mention}" for react, r in found))
            parsed_found = tuple((e, r.id) for e, r in found)
            for r, _ in found:
                try:
                    await message.add_reaction(r)
                except discord.HTTPException:
                    await message.delete()
                    raise CommandSyntaxError('Do not use emoji unavailable to the bot.')
            self.reacts[str(message.id)] = {
                "type": MESSAGE_TYPE.CHOICE if args["exclusive"] else MESSAGE_TYPE.NORMAL,
                "required": requisite,
                "reacts": parsed_found
            }
            self.storage_file.save()

    @Command("PasswordRole", "PWDRole",
             doc="Specifies a role to be granted to people offering a correct password through ClaimRole command.\n"
                 "Pass phrases with spaces in them can be supplied in quotation marks, e.g. \"some pass phrase\".\n"
                 "NOTE: whitespace will be reduced to single space to prevent confusion.\n"
                 "If a given pass phrase already exists on the server, role will be updated or deleted if no role is "
                 "supplied.\n"
                 "If no arguments are given, lists existing password-role bindings.",
             syntax="[password] [role]",
             perms={"manage_roles"},
             category="role_request")
    async def _password_role(self, msg: discord.Message):
        args = shlex.split(msg.content)
        args_len = len(args)

        if args_len == 1:  # no arguments given, return list. There can never be an empty args list, because command.
            for message in group_items(
                    [f"{key}: {msg.guild.get_role(role['role']).name}" for key, role in self.passwords.items()],
                    message="**ANALYSIS: Following pass phrase - role pairs found:**\n",
            ):
                await respond(msg, message)
            return

        role = None
        if args_len == 3:  # optional role parameter detected. Not joining it up ourselves
            role_query = args[2]
            role = find_role(msg.guild, role_query)
            if not role:
                raise CommandSyntaxError(f"Role `{role_query}` not found. Try using a mention or a role id.")

        #  clean up password white space, to have mercy on our users.
        #  if you think that filtering out reading capability with variable spacing in password is a good idea...
        #  well you are wrong.
        password = " ".join(args[1].split())

        if password not in self.passwords:  # add new password - role pair
            self.passwords[password] = {
                "role": role.id
            }
            await respond(msg, f"**AFFIRMATIVE. Added pass phrase `{password}`, granting role `{role.name}`.**")
        elif role:  # replace role
            self.passwords[password]["role"] = role.id
            await respond(msg, f"**AFFIRMATIVE. Pass phrase `{password}` now grants role `{role.name}`.**")
        else:  # delete password - role pair
            del self.passwords[password]
            await respond(msg, f"**AFFIRMATIVE. Pass phrase `{password}` has been removed.**")
        self.storage_file.save()

    @Command("ClaimRole",
             syntax="(password)",
             category="role_request",
             run_anywhere=True,
             delete_call=True)
    async def _claim_role(self, msg: discord.Message):
        _, *password = msg.content.split()
        if not password:
            return

        password = " ".join(password)

        if password not in self.passwords:
            return

        role = msg.guild.get_role(self.passwords[password]['role'])
        if not role:
            return

        await msg.author.add_roles(role)

    async def on_raw_reaction_add(self, payload: discord.RawReactionActionEvent):
        """
        :param payload:
        :return:
        """
        if payload.user_id == self.client.user.id:
            return

        message_id = str(payload.message_id)

        if message_id not in self.reacts:
            return

        offering = self.reacts[message_id]

        roles = []
        msg = await self.client.get_channel(payload.channel_id).fetch_message(payload.message_id)
        user = self.client.get_guild(payload.guild_id).get_member(payload.user_id)
        emoji = str(payload.emoji)

        if offering["required"]:
            for required_role_id in offering["required"]:
                if not user.get_role(required_role_id):
                    await msg.remove_reaction(payload.emoji, user)
                    return

        if offering["type"] == MESSAGE_TYPE.CHOICE:
            to_add = set()
            to_remove = set()
            to_remove_emoji = set()
            for reaction, roleId in offering["reacts"]:
                if reaction == emoji:
                    to_add.add(roleId)
                else:
                    to_remove.add(roleId)
                    to_remove_emoji.add(reaction)
            to_remove -= to_add
            roles = [msg.guild.get_role(role_id) for role_id in to_add]
            if to_remove:
                await asyncio.gather(*(msg.remove_reaction(emoji, user) for emoji in to_remove_emoji))
                await user.remove_roles(*(msg.guild.get_role(role_id) for role_id in to_remove),
                                        reason="Removed by request through plugin.")
        else:
            roles = [msg.guild.get_role(role_id) for reaction, role_id in offering["reacts"] if reaction == emoji]

        if roles:
            await user.add_roles(*roles, reason="Added by request through plugin.")
            return
        await msg.remove_reaction(payload.emoji, user)

    async def on_raw_reaction_remove(self, payload: discord.RawReactionActionEvent):
        """
        :param payload:
        :return:
        """
        if payload.user_id == self.client.user.id:
            return

        message_id = str(payload.message_id)

        if message_id not in self.reacts:
            return

        offer = self.reacts[message_id]
        msg = await self.client.get_channel(payload.channel_id).fetch_message(payload.message_id)
        user = self.client.get_guild(payload.guild_id).get_member(payload.user_id)
        emoji = str(payload.emoji)

        roles = [msg.guild.get_role(role_id) for reaction, role_id in offer["reacts"] if reaction == emoji]

        if roles:
            await user.remove_roles(*roles, reason="Removed by request through plugin.")
            return

    async def on_message_delete(self, msg: discord.Message):
        mid = str(msg.id)
        if mid in self.reacts:
            del self.reacts[mid]
            self.storage_file.save()



from __future__ import annotations
from red_star.plugin_manager import BasePlugin
from red_star.rs_utils import respond, find_user, is_positive, group_items, prompt_for_confirmation
from red_star.command_dispatcher import Command
from red_star.rs_errors import CommandSyntaxError
import discord
import json


class Levelling(BasePlugin):
    name = "levelling"
    version = "1.1"
    author = "GTG3000"
    description = "A plugin for providing an XP system that awards members XP for messages."
    default_config = {
        "low_cutoff": 75,
        "xp_min": 1,
        "xp_max": 10,
        "skip_missing": False
    }
    channel_categories = {"no_xp"}

    async def activate(self):
        self._port_old_storage()
        self.storage.setdefault("xp", {})

    def _port_old_storage(self):
        old_storage_path = self.config_manager.config_path / "xp.json"
        if old_storage_path.exists():
            with old_storage_path.open(encoding="utf-8") as fp:
                old_storage = json.load(fp)
            for guild_id, xp_storage in old_storage.items():
                try:
                    new_storage = self.config_manager.storage_files[guild_id][self.name]
                except KeyError:
                    self.logger.warn(f"Server with ID {guild_id} not found! Is the bot still in this server?\n"
                                     f"Skipping conversion of this server's XP storage...")
                    continue
                new_storage.contents["xp"] = xp_storage
                new_storage.save()
                new_storage.load()
            old_storage_path = old_storage_path.replace(old_storage_path.with_suffix(".json.old"))
            self.logger.info(f"Old XP storage converted to new format. "
                             f"Old data now located at {old_storage_path} - you may delete this file.")

    async def on_message(self, msg: discord.Message):
        if not self.channel_manager.channel_in_category("no_xp", msg.channel):
            self._give_xp(msg)

    async def on_message_delete(self, msg: discord.Message):
        if not self.channel_manager.channel_in_category("no_xp", msg.channel):
            self._take_xp(msg)

    # Commands

    @Command("ListXP", "XPLeaderboard",
             doc="Lists all registered users from highest XP to lowest, up to amount specified or 10.",
             syntax="[number]",
             category="levelling")
    async def _list_xp(self, msg: discord.Message):
        skip = self.config['skip_missing']

        args = msg.content.split()
        if len(args) > 1:
            try:
                limit = int(args[1])
            except ValueError:
                raise CommandSyntaxError(f"{args[1]} is not a valid integer.")
        else:
            limit = 10

        # Iterate over all the xp values, counting up the position and forming a nice string.
        # Not using enumerate because some positions may be skipped due to missing members.
        # Could possibly squeeze it into a list comprehension, but it would be a monstrosity.

        pos = 1
        xp_list = []
        for uid in sorted(self.storage["xp"], key=self.storage["xp"].get, reverse=True):
            user = msg.guild.get_member(int(uid))

            if user:
                user = user.display_name
            elif skip:
                continue
            else:
                user = uid

            xp_list.append(f"{pos:03d}|{user:<32}|{self.storage['xp'][uid]:>6}")
            pos += 1

            if pos > limit:
                break

        if xp_list:
            for string in group_items(xp_list, message="**ANALYSIS: Current XP leaderboard:**"):
                await respond(msg, string)

    @Command("XP", "ShowXP",
             doc="Shows your xp or xp of specified user.",
             syntax="[user]",
             category="levelling")
    async def _xp(self, msg: discord.Message):
        args = msg.content.split(None, 1)

        if len(args) > 1:
            user = find_user(msg.guild, args[1])
            if user:
                if str(user.id) in self.storage["xp"]:
                    await respond(msg, f"**ANALYSIS: User {user.display_name} has "
                                       f"{self.storage['xp'][str(user.id)]} XP.**")
                else:
                    await respond(msg, f"**WARNING: User {user.display_name} has no XP record.**")
            else:
                raise CommandSyntaxError("Not a user or user not found.")
        else:
            uid = str(msg.author.id)
            if uid in self.storage["xp"]:
                await respond(msg, f"**ANALYSIS: You have {self.storage['xp'][uid]} XP.**")
            else:
                await respond(msg, "**ANALYSIS: You have no XP record.**")  # I don't think this is possible

    @Command("EvalXP",
             doc="Processes all message history and grants members xp.\nAccepts one argument to determine "
                 "how far into the history to search.\nWARNING - VERY SLOW.\nUSE ONLY AFTER "
                 "CLEANING XP TABLE.",
             syntax="[depth]",
             perms={"manage_guild"},
             category="levelling")
    async def _eval_xp(self, msg: discord.Message):
        args = msg.content.split(None, 1)
        if len(args) > 1:
            try:
                depth = int(args[1])
            except ValueError:
                raise CommandSyntaxError("Argument is not a valid integer.")
        else:
            depth = None

        display = await respond(msg, "**AFFIRMATIVE. Processing messages.**")
        async with msg.channel.typing():
            for channel in self.guild.text_channels:
                if not self.channel_manager.channel_in_category("no_xp", channel):
                    await display.edit(content=f"**AFFIRMATIVE. Processing messages in channel {channel}.**")
                    try:
                        async for message in channel.history(limit=depth):
                            self._give_xp(message)
                    except discord.Forbidden:
                        continue

        self.storage_file.save()
        await display.delete()

    @Command("NukeXP",
             doc="Permanently erases XP records, setting given user or EVERYONE to 0.",
             syntax="[user]",
             perms={"manage_guild"},
             category="levelling")
    async def _nuke_xp(self, msg: discord.Message):
        args = msg.content.split(None, 1)

        if len(args) > 1:
            user = find_user(msg.guild, args[1])
            if user:
                if str(user.id) in self.storage["xp"]:
                    del self.storage["xp"][str(user.id)]
                    await respond(msg, f"**AFFIRMATIVE. User {user.display_name} was removed from XP table.**")
                else:
                    await respond(msg, f"**NEGATIVE. User {user.display_name} has no XP record.**")
            elif args[1] in self.storage["xp"]:
                del self.storage["xp"][args[1]]
                await respond(msg, f"**AFFIRMATIVE. ID {args[1]} was removed from XP table.**")
            else:
                raise CommandSyntaxError("Not a user or no user found.")
        else:
            prompt_text = f"This will destroy ALL XP records for the entire server. Continue?"
            confirmed = await prompt_for_confirmation(msg, prompt_text=prompt_text)
            if not confirmed:
                return
            self.storage["xp"] = {}
            self.storage_file.save()
            await respond(msg, "**AFFIRMATIVE. XP table deleted.**")

    @Command("XPConfig", "XPSettings",
             doc="Edit the xp module settings, or see the current settings.\n"
                 "After adjusting these settings, it is advisable to recalculate server XP values by running NukeXP "
                 "and EvalXP.",
             syntax="[option] [value]",
             perms={"manage_guild"},
             category="levelling")
    async def _set_xp(self, msg: discord.Message):
        args = msg.content.split(" ", 2)
        cfg = self.config
        if len(args) == 1:
            missing_member_str = ("skipping" if cfg["skip_missing"] else "displaying") + " missing members"
            await respond(msg, "**ANALYSIS: Current XP settings:**```\n"
                               f"low_cutoff: {cfg['low_cutoff']}\n"
                               f"xp_min    : {cfg['xp_min']}\n"
                               f"xp_max    : {cfg['xp_max']}\n"
                               f"missing   : {missing_member_str}```")
        elif len(args) == 3:
            if args[1].lower() == "missing":
                cfg["skip_missing"] = is_positive(args[2])
            else:
                try:
                    val = int(args[2])
                except ValueError:
                    raise CommandSyntaxError("Second argument must be an integer.")
                else:
                    if args[1].lower() == "low_cutoff":
                        cfg["low_cutoff"] = val
                    elif args[1].lower() == "xp_min":
                        cfg["xp_min"] = val
                    elif args[1].lower() == "xp_max":
                        cfg["xp_max"] = val
                    else:
                        raise CommandSyntaxError(f"No option {args[1].lower()}")
        else:
            raise CommandSyntaxError("Two arguments required.")

    # Utilities

    def _give_xp(self, msg: discord.Message):
        uid = str(msg.author.id)
        if uid in self.storage["xp"]:
            self.storage["xp"][uid] += self._calc_xp(msg.clean_content)
        else:
            self.storage["xp"][uid] = self._calc_xp(msg.clean_content)

    def _take_xp(self, msg: discord.Message):
        uid = str(msg.author.id)
        if uid in self.storage["xp"]:
            self.storage["xp"][uid] -= self._calc_xp(msg.clean_content)

    def _calc_xp(self, txt: str):
        """
        :type txt:str
        :param txt:message to calculate the xp for
        :return:
        """
        spc = self.config

        if len(txt) < spc["low_cutoff"]:
            return 0

        t_percent = (len(txt)-spc["low_cutoff"])/(2000-spc["low_cutoff"])

        t_xp = spc["xp_min"] + (spc["xp_max"]-spc["xp_min"])*t_percent

        return int(t_xp)



import discord
from red_star.plugin_manager import BasePlugin
from red_star.rs_errors import ChannelNotFoundError
from red_star.rs_utils import sub_user_data, respond
from random import choice


class Announcer(BasePlugin):
    name = "announcer"
    version = "1.1"
    author = "medeor413"
    description = "A plugin for basic automated announcements, triggering messages when the bot starts, " \
                  "a new member joins, or the bot is pinged."
    default_config = {
        "greeting_message": "**RED STAR IS ONLINE.**",
        "new_member_announce_message": "**Welcome to the server, <@usermention>.**",
        "ping_messages": True,
        "ping_messages_on_everyone": False,
        "ping_message_options": [
            "**NEGATIVE**",
            "**AFFIRMATIVE**",
            "**WARNING: Does not compute.**",
            "**ANALYSIS: Maybe.**"
        ]
    }
    channel_types = {"startup", "welcome"}

    async def on_all_plugins_loaded(self):
        await self._greet()

    async def _greet(self):
        try:
            greet_channel = self.channel_manager.get_channel("startup")
            await greet_channel.send(self.config["greeting_message"])
        except ChannelNotFoundError:
            pass

    async def _ping_response(self, msg: discord.Message):
        response = sub_user_data(msg.author, choice(self.config["ping_message_options"]))
        await respond(msg, response)

    # Event hooks

    async def on_message(self, msg: discord.Message):
        ping_messages = self.config["ping_messages"]
        ping_messages_everyone = self.config["ping_messages_on_everyone"]
        if ping_messages and (ping_messages_everyone >= msg.mention_everyone) and msg.guild.me.mentioned_in(msg):
            await self._ping_response(msg)

    async def on_member_join(self, member: discord.Member):
        text = sub_user_data(member, self.config["new_member_announce_message"])
        try:
            chan = self.channel_manager.get_channel("welcome")
            await chan.send(text)
        except ChannelNotFoundError:
            pass



from __future__ import annotations
import discord
import inspect
import logging
import importlib
import importlib.util
from sys import exc_info, modules
from types import ModuleType
from red_star.channel_manager import ChannelManager
from red_star.command_dispatcher import CommandDispatcher

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pathlib import Path
    from typing import Type
    from red_star.client import RedStar
    from red_star.config_manager import ConfigManager, PluginStorageFile, JsonValues


class PluginManager:
    """
    Manages the loading of plugins and dispatching of event hooks.
    """

    def __init__(self, client: RedStar):
        self.client = client
        self.config_manager = client.config_manager
        self.logger = logging.getLogger("red_star.plugin_manager")
        self.default_server_config = {"disabled_plugins": []}
        self.last_error = None

        self.modules: dict[str, ModuleType] = dict()
        self.plugin_classes: dict[str, Type[BasePlugin]] = dict()

        self.plugins: dict[discord.Guild, dict[str, BasePlugin]] = dict()
        self.command_dispatchers: dict[discord.Guild, CommandDispatcher] = dict()
        self.channel_managers: dict[discord.Guild, ChannelManager] = dict()
        self.plugin_package = ModuleType("red_star_plugins")
        self.plugin_package.__path__ = []
        modules["red_star_plugins"] = self.plugin_package

        self.logger.debug("Initialized plugin manager.")

    def __repr__(self):
        return f"<PluginManager: Plugins: {self.plugin_classes.keys()}>"

    def load_all_plugins(self, plugin_paths: list[Path]):
        """
        Loads all plugins from all specified folders and places them into the `red_star_plugins` module for
        cross-plugins access.
        :param plugin_paths:
        :return:
        """
        self.logger.debug("Loading plugins...")
        self.plugin_package.__path__.extend(str(x) for x in plugin_paths)
        for path in plugin_paths:
            self._load_plugin_folder(path)
        self.config_manager.save_config()
        self.logger.info(f"Loaded {len(self.plugin_classes)} plugins from {len(self.modules)} modules.")

    def _load_plugin_folder(self, plugin_path: Path):
        """
        Loads all plugins in a folder and initializes them for use.
        :param plugin_path:
        :return:
        """
        self.logger.debug(f"Loading plugins from {plugin_path}...")
        loaded = set()
        plugin_path.mkdir(parents=True, exist_ok=True)
        for file in plugin_path.iterdir():
            if file.stem.startswith(("_", ".")):
                continue
            if (file.suffix == ".py" or file.is_dir()) and file not in loaded:
                try:
                    modul = self._load_module(file.stem)
                    self.load_plugin(modul)
                    loaded.add(file)
                except (SyntaxError, ImportError):
                    self.logger.exception(f"Exception encountered loading plugin {file.stem}: ", exc_info=True)
                    continue
                except FileNotFoundError:
                    self.logger.error(f"File {file.stem} missing when load attempted!")
                    continue

    def _load_module(self, module_name: str) -> ModuleType:
        """
        Imports Python modules containing plugins and adds them to the module list.

        :param module_name: the name of the module to be imported.
        :return: The module object.
        """
        mod = importlib.import_module(f"red_star_plugins.{module_name}")
        self.modules[module_name] = mod
        self.logger.debug(f"Imported module {module_name}.")
        return mod

    def _get_plugin_class(self, plugin_module: ModuleType) -> set[Type[BasePlugin]]:
        """
        Extracts the plugin classes from a module and assigns them several class-level properties.

        :param plugin_module: The module containing the plugin classes to be extracted.
        :return: The extracted plugin classes, with class properties assigned.
        """
        def predicate(cls):
            return inspect.isclass(cls) and issubclass(cls, BasePlugin) and cls is not BasePlugin

        class_list = set()
        for name, obj in inspect.getmembers(plugin_module, predicate=predicate):
            obj.client = self.client
            obj.config_manager = self.config_manager
            obj.global_plugin_config = self.config_manager.get_global_config(obj.name,
                                                                             default_config=obj.default_global_config)
            obj.plugin_manager = self
            ChannelManager.channel_types.update(obj.channel_types)
            ChannelManager.channel_categories.update(obj.channel_categories)
            class_list.add(obj)
        return class_list

    def load_plugin(self, plugin_module: ModuleType):
        """
        Extracts the plugin classes from a module and creates instances of them for use,
        placing them in the plugins list.
        :param plugin_module: The module containing the plugin classes to be extracted.
        """
        classes = self._get_plugin_class(plugin_module)
        for cls in classes:
            self.plugin_classes[cls.name] = cls
            self.logger.debug(f"Loaded plugin {cls.name}")

    async def activate_all(self):
        """
        Activates enabled plugins by running their activate() function and placing them in the active plugins list.
        Runs the on_all_plugins_loaded hook after loading all plugins.
        :return:
        """
        self.logger.info("Activating plugins.")
        for guild in self.client.guilds:
            await self.activate_server_plugins(guild)

    async def activate_server_plugins(self, guild: discord.Guild):
        channel_manager = ChannelManager(self.client, guild)
        self.channel_managers[guild] = channel_manager
        command_dispatcher = CommandDispatcher(self.client, guild, channel_manager)
        self.command_dispatchers[guild] = command_dispatcher
        self.plugins[guild] = {}
        disabled_plugins = self.config_manager.get_server_config(guild, "plugin_manager",
                                                                 self.default_server_config)["disabled_plugins"]

        for name, plugin in self.plugin_classes.items():
            if name in disabled_plugins:
                continue
            await self.activate(guild, name)
        await self.hook_event("on_all_plugins_loaded", guild)

    async def activate(self, guild: discord.Guild, name: str):
        guild_plugins = self.plugins[guild]
        try:
            plugin = self.plugin_classes[name]
            if name not in guild_plugins:
                self.logger.info(f"Activating plugin {name}.")
                # noinspection PyBroadException
                try:
                    plugin_inst = plugin(guild,
                                         self.config_manager.get_server_config(guild, name, plugin.default_config),
                                         self.channel_managers[guild], self.command_dispatchers[guild], guild_plugins)
                    plugin_inst.storage_file = self.config_manager.get_plugin_storage(plugin_inst)
                    await plugin_inst.activate()
                    self.command_dispatchers[guild].register_plugin(plugin_inst)
                    guild_plugins[name] = plugin_inst
                except Exception:
                    self.logger.exception(
                        f"Error occurred while activating plugin {plugin.name} for server {guild.id}: ",
                        exc_info=True)
            else:
                self.logger.warning(f"Attempted to activate already active plugin {name}.")
        except KeyError:
            self.logger.error(f"Attempted to activate non-existent plugin {name}.")

    async def deactivate_all(self):
        """
        Deactivates all enabled plugins, typically in preparation for shutdown. Removes them from the active plugins
        list and runs their deactivate() function.
        :return:
        """
        self.logger.info("Deactivating plugins.")
        for guild in self.plugins.keys():
            await self.deactivate_server_plugins(guild)

    async def deactivate_server_plugins(self, guild):
        guild_plugins = self.plugins[guild]
        for name, plugin in guild_plugins.copy().items():
            await self.deactivate(guild, name)
        del self.command_dispatchers[guild]
        del self.channel_managers[guild]

    async def deactivate(self, guild: discord.Guild, name: str):
        guild_plugins = self.plugins[guild]
        try:
            if name in guild_plugins:
                plugin = guild_plugins[name]
                self.logger.info(f"Deactivating plugin {name}.")
                # noinspection PyBroadException
                try:
                    await plugin.deactivate()
                except Exception:
                    self.logger.exception(f"Error occurred while deactivating plugin {name}: ", exc_info=True)
                self.command_dispatchers[guild].deregister_plugin(plugin)
                del guild_plugins[name]
                await self.hook_event("on_plugin_deactivated", name)
            else:
                self.logger.warning(f"Attempted to deactivate already inactive plugin {name}.")
        except KeyError:
            self.logger.error(f"Attempted to deactivate non-existent plugin {name}.")

    async def reload_plugin(self, name: str):
        """
        Reloads a plugin from its source file.
        :param name: The plugin to be reloaded.
        """
        try:
            self.logger.info(f"Reloading plugin module {name}.")
            servers_active_on = set()
            for guild, plugins in self.plugins.items():
                if name in plugins:
                    servers_active_on.add(guild)
                    await self.deactivate(guild, name)
            del self.plugin_classes[name]
            modul = self.modules[name]
            importlib.reload(modul)
            self.load_plugin(modul)
            for guild in servers_active_on:
                await self.activate(guild, name)
        except KeyError:
            self.logger.error(f"Attempted to reload non-existent plugin module {name}.")

    async def hook_event(self, event: str, guild, *args, **kwargs):
        """
        Dispatches an event, with its data, to all plugins.

        :param event: The name of the event. Should match the calling function.
        :param guild: The guild to which the event belongs.
        :param args: Everything that gets passed to the calling function
        should be passed through to this function.
        """
        if event == "on_message" and guild in self.command_dispatchers:
            await self.command_dispatchers[guild].on_message(*args)

        guild_plugins = set(self.plugins.get(guild, {}).values())
        for plugin in guild_plugins:
            hook = getattr(plugin, event, None)
            if hook:
                # noinspection PyBroadException
                try:
                    await hook(*args, **kwargs)
                except Exception:
                    self.last_error = exc_info()
                    self.logger.exception(f"Exception encountered in plugin {plugin.name} on event {event}: ",
                                          exc_info=True)


class BasePlugin:
    """
    The base plugin class from which all plugin classes must inherit in order to be detected by the plugin manager.
    It is recommended you set all meta-fields (name, description, version, author), but only name is necessary.
    In most circumstances you're going to want the plugin name and the module name to be the same.
    The plugin manager will install various useful things to the class.
    """
    # Metadata fields
    name: str = "Base Plugin"
    description: str = "This is a template class for plugins. Name *must* be filled, other meta-fields are optional."
    version: str = "1.0"
    author: str = "Unknown"

    # Attributes added by plugin manager
    client: RedStar
    config_manager: ConfigManager
    plugin_manager: PluginManager

    # Developer-defined attributes for use in code
    default_config: dict = {}
    default_global_config: dict = {}
    global_plugin_config: dict = {}
    channel_types: set = set()
    channel_categories: set = set()

    def __init__(self, guild: discord.Guild, plugin_config: dict, channel_manager: ChannelManager,
                 command_dispatcher: CommandDispatcher, plugins: dict[str, BasePlugin]):
        self.guild = guild
        self.config = plugin_config
        self.channel_manager = channel_manager
        self.command_dispatcher = command_dispatcher
        self.plugins = plugins
        self.logger = logging.getLogger(f"red_star.plugin.{self.name}.{guild.id}")
        self.storage_file: PluginStorageFile

    @property
    def storage(self) -> JsonValues:
        return self.storage_file.contents

    @storage.setter
    def storage(self, value: JsonValues):
        self.storage_file.contents = value

    def storage_load_args(self):
        return {}

    def storage_save_args(self):
        return {}

    async def activate(self):
        """
        The method called when the plugin is initialized. Should be used to get all the Discord-related
        initialization out of the way.
        Raise an exception in this method to cancel activation, say if a required package isn't installed.
        """

    async def deactivate(self):
        """
        The method called when the plugin is uninitialized. Should be used to perform any necessary cleanup.
        """

    def __str__(self):
        """
        Method to return something a little less nasty.
        :return: string: The string to return when str() is called on this object.
        """
        return f"<Plugin {self.name} (Version {self.version}) for server {self.guild.id}>"

    def __repr__(self):
        """
        Method to return something a little less nasty.
        :return: string: The string to return when repr() is called on this object.
        """
        return f"<Plugin {self.name} (Version {self.version}) for server {self.guild.id}>"



from __future__ import annotations
import re
import json
import shlex
import discord
from red_star.rs_errors import CommandSyntaxError, UserPermissionError
from red_star.rs_utils import respond, find_role, find_user, split_message, decode_json, RSArgumentParser
from red_star.command_dispatcher import Command
from red_star.plugin_manager import BasePlugin
from io import BytesIO
from dataclasses import dataclass, asdict


class Roleplay(BasePlugin):
    name = "roleplay"
    version = "1.2"
    author = "GTG3000"

    default_config = {
        "allow_race_requesting": False,
        "race_roles": [],
        "pinned_bios": {},
        "pinned_bios_channel": False
    }

    @dataclass
    class Bio:
        author: int
        name: str
        race: str
        gender: str
        appearance: str
        backstory: str
        height: str
        age: str
        theme: str
        link: str
        image: str
        equipment: str
        skills: str
        personality: str
        interests: str

        fields = ["author", "name", "race", "gender", "height", "age", "theme", "link", "image", "appearance",
                  "equipment", "skills", "personality", "backstory", "interests"]
        mandatory_fields = ["name", "race", "gender", "appearance", "backstory"]

        lim_64 = ["race", "gender", "height", "age", "name"]

        def set(self, field: str, value: str = None):
            _field = field.lower()
            if _field not in self.fields:
                raise KeyError(f"{_field} is not a valid field.")
            if value:
                if len(value) > (64 if _field in self.lim_64 else 1024):
                    raise ValueError('64' if _field in self.lim_64 else '1024')
                self.__dict__[_field] = self._name(value) if _field == 'name' else value
            else:
                self.__dict__[_field] = 'undefined' if _field in self.mandatory_fields else ''

        @classmethod
        def blank_bio(cls, author: int, name: str) -> Roleplay.Bio:
            new_bio_dict = dict(zip(cls.fields, [''] * 15))
            for field in cls.mandatory_fields:
                new_bio_dict[field] = 'undefined'

            new_bio_dict['author'] = author
            new_bio_dict['name'] = cls._name(name)

            return cls(**new_bio_dict)

        @staticmethod
        def _name(name: str) -> str:
            """
            removes trailing/leading whitespace, limits whitespace between words to one space, removes newlines.
            :param name:
            :return:
            """
            clean = re.sub(r'\s+', ' ', re.sub(r'^\s+|\s+$|\n|\r', '', name))
            if not clean:
                raise CommandSyntaxError('Empty name provided.')
            return clean[:64]

        def embed(self, guild: discord.Guild, roles: [discord.Role]) -> discord.Embed:
            """
            Generates a pretty discord embed of this role.

            :param guild: guild that the bio belongs to, for member and role searching
            :param roles: list of accepted race roles
            :return: Embed object representing role
            """
            t_embed = discord.Embed(type="rich", colour=16711680)

            role = find_role(guild, self.race)

            t_embed.title = self.name

            if role and role.id in roles:
                t_embed.colour = role.colour

            t_member = guild.get_member(self.author)
            if t_member:
                t_embed.set_footer(text=f"Character belonging to {t_member.display_name}",
                                   icon_url=t_member.avatar.url)

            t_embed.description = "```\n" + \
                                  '\n'.join([f"{f.capitalize():<7}: {self.__dict__[f]}" for f in self.fields[2:6] if
                                             self.__dict__[f]]) \
                                  + "```\n" + \
                                  (f"[Theme song.]({self.theme})\n" if self.theme else '') + \
                                  (f"[Extended bio.]({self.link})\n" if self.link else '') + \
                                  (f"Owner: {t_member.mention}" if t_member else '')

            if self.image:
                t_embed.set_image(url=self.image)

            for field in (f for f in self.fields[9:] if self.__dict__[f]):
                t_embed.add_field(name=field.capitalize(), value=self.__dict__[field], inline=False)

            return t_embed

        def as_dict(self) -> dict:
            return {"__classhint__": "bio", **asdict(self)}

    def storage_load_args(self):
        return {"object_hook": self._load_bio}

    def storage_save_args(self):
        return {"default": lambda obj: obj.as_dict(), "indent": 2, "ensure_ascii": False}

    def _load_bio(self, obj: dict) -> Roleplay.Bio | dict:
        if obj.pop('__classhint__', None) == 'bio':
            return self.Bio(**obj)
        else:
            obj_set = {*obj}
            if obj_set.issubset(self.Bio.fields) and len(obj_set) >= 5:
                return self.Bio(**{**dict(zip(self.Bio.fields, [''] * 15)), **obj})
            else:
                return obj

    async def activate(self):
        self._port_old_storage()
        self.bios = self.storage.setdefault("bios", {})

    def _port_old_storage(self):
        old_storage_path = self.config_manager.config_path / "bios.json"
        if old_storage_path.exists():
            with old_storage_path.open(encoding="utf-8") as fp:
                old_storage = json.load(fp, **self.storage_load_args())
            for guild_id, data in old_storage.items():
                old_storage_guild = old_storage.get(guild_id, {})
                try:
                    guild_current_storage = self.config_manager.storage_files[guild_id][self.name]
                except KeyError:
                    self.logger.warn(f"Server with ID {guild_id} not found! Is the bot still in this server?\n"
                                     f"Skipping conversion of this server's bio storage...")
                    continue
                old_storage_guild.update(guild_current_storage.contents)
                guild_current_storage.contents = old_storage_guild
                guild_current_storage.save()
                guild_current_storage.load()
            old_storage_path = old_storage_path.replace(old_storage_path.with_suffix(".json.old"))
            self.logger.info(f"Old bio storage converted to new format. "
                             f"Old data now located at {old_storage_path} - you may delete this file.")

    @Command("RaceRole",
             doc="-a/--add   : Adds specified roles to the list of allowed race roles.\n"
                 "-r/--remove: Removes specified roles from the list.\n"
                 "Calling it without any arguments prints the list.",
             syntax="[-a/--add (role mentions/ids/names)] [-r/--remove (role mentions/ids/names)]",
             perms={"manage_messages"},
             category="role_play")
    async def _race_role(self, msg: discord.Message):
        parser = RSArgumentParser()
        parser.add_argument("command")
        parser.add_argument("-a", "--add", default=[], nargs='+')
        parser.add_argument("-r", "--remove", default=[], nargs='+')

        args = parser.parse_args(shlex.split(msg.content))

        if not (args['add'] or args['remove']):
            approved_roles = "\n".join(x.name for x in msg.guild.roles if x.id in self.config["race_roles"])
            for split_msg in split_message(f"**ANALYSIS: Currently approved race roles:**```\n{approved_roles}```"):
                await respond(msg, split_msg)
        else:
            args['add'] = [r for r in [find_role(msg.guild, r) for r in args['add']] if r]
            args['remove'] = [r for r in [find_role(msg.guild, r) for r in args['remove']] if r]

            # for nice output
            added_roles = []
            removed_roles = []

            for role in args['add']:
                if role.id not in self.config["race_roles"]:
                    added_roles.append(role.name)
                    self.config["race_roles"].append(role.id)
            for role in args['remove']:
                if role.id in self.config["race_roles"]:
                    removed_roles.append(role.name)
                    self.config["race_roles"].remove(role.id)

            if added_roles or removed_roles:
                output_str = "**AFFIRMATIVE. ANALYSIS:**\n```diff\n"
                if added_roles:
                    output_str += "Added roles:\n+ " + "\n+ ".join(added_roles) + "\n"
                if removed_roles:
                    output_str += "Removed roles:\n- " + "\n- ".join(removed_roles) + "\n"
                output_str += "```"
                await respond(msg, output_str)
            else:
                raise CommandSyntaxError

    @Command("GetRaceRole",
             doc="Allows the user to request one of the approved race roles for themselves.",
             syntax="(role)",
             category="role_play")
    async def _get_race_role(self, msg: discord.Message):
        if not self.config.get("allow_race_requesting", False):
            return
        args = msg.content.split(" ", 1)
        preexisting_roles = []
        for role in msg.author.roles:
            if role.id in self.config["race_roles"]:
                preexisting_roles.append(role)
        await msg.author.remove_roles(*preexisting_roles)
        if len(args) < 2:
            await respond(msg, "**AFFIRMATIVE. Race role removed.**")
        else:
            role = find_role(msg.guild, args[1])
            if role:
                if role.id in self.config["race_roles"]:
                    await msg.author.add_roles(role)
                    await respond(msg, f"**AFFIRMATIVE. Race role {role.name} granted.**")
                else:
                    raise CommandSyntaxError("Not an approved race role.")
            else:
                raise CommandSyntaxError("Not a role or role not found.")

    @Command("ListRaceRoles",
             doc="Lists all approved race roles.",
             category="role_play")
    async def _list_race_roles(self, msg: discord.Message):
        if not self.config.get("allow_race_requesting", False):
            return
        approved_roles = "\n".join(x.name for x in msg.guild.roles if x.id in self.config["race_roles"])
        for split_msg in split_message(f"**ANALYSIS: Currently approved race roles:**```\n{approved_roles}```"):
            await respond(msg, split_msg)

    @Command("ListBios",
             doc="Lists all available bios in the database.",
             syntax="[user]",
             category="role_play")
    async def _list_bios(self, msg: discord.Message):
        args = msg.content.split(" ", 1)
        if len(args) > 1:
            owner = find_user(msg.guild, args[1])
            if owner:
                result = "\n".join(f"{k[:16]:<16} : {v.name}" for k, v in self.bios.items()
                                   if v.author == owner.id)
                for split_msg in split_message(f"**ANALYSIS: User {owner.display_name} has following "
                                               f"characters:**```{result}```"):
                    await respond(msg, split_msg)
            else:
                raise CommandSyntaxError("Not a user or user not found.")
        else:
            bios = "\n".join(f"{k[:16]:<16} : {v.name}" for k, v in self.bios.items())
            for split_msg in split_message(f"**ANALYSIS: Following character bios found:**```\n{bios}```"):
                await respond(msg, split_msg)

    @Command("Bio",
             doc="Prints, edits, creates, destroys, renames or dumps character bios.\n"
                 "Each character name must be unique and will be stripped of excessive whitespace.\n"
                 "-s/--set   : changes a specified field to a specified value:\n"
                 "  Fields   : name/race/gender/height/age: limit 64 characters."
                 "theme/link: must be viable http(s) url. "
                 "  appearance/equipment/skills/personality/backstory/interests: limit 1024 characters.\n"
                 "  Setting 'race' to the same name as a registered character role will fetch the colour.\n"
                 "-c/--create: creates a new bio with the given name.\n"
                 "-d/--dump  : creates and uploads a JSON file of the bio, for backup and offline editing.\n"
                 "-r/--rename: changes the name by which the bio is referenced to a new one.\n"
                 "--delete   : deletes the bio."
                 "Be aware that the total length of the bio must not exceed 6000 characters.",
             syntax="(name) [-s/--set (field) [value]] [-c/--create] [-d/--dump] [-r/--rename (new name)] [--delete]",
             category="role_play",
             run_anywhere=True,
             optional_perms={"edit_others": {"manage_messages"}})
    async def _bio(self, msg: discord.Message):
        parser = RSArgumentParser()
        parser.add_argument('command')
        parser.add_argument('name', default=[], nargs="*")
        parser.add_argument('-s', '--set', default=[], nargs="+")
        parser.add_argument('-c', '--create', action='store_true')
        parser.add_argument('-d', '--dump', action='store_true')
        parser.add_argument('--delete', action='store_true')  # doesn't get a short flag for safety
        parser.add_argument('-r', '--rename', nargs="+")

        try:
            args = shlex.split(msg.clean_content)
        except ValueError as e:
            self.logger.warning("Unable to split {data.content}. {e}")
            raise CommandSyntaxError(e)

        args = parser.parse_args(args)

        if args['name']:
            args['name'] = self.Bio._name(' '.join(args['name']))
            char = args['name'].lower()
        else:
            raise CommandSyntaxError("No bio id given.")

        if char not in self.bios and not args['create']:
            raise CommandSyntaxError(f'No such character: {args["name"]}.')

        # manipulate the specified bio
        if args['set'] or args['create'] or args['delete'] or args['rename'] or args['dump']:
            # creating a bio with using the given character name
            if args['create']:
                if char in self.bios:
                    raise CommandSyntaxError(f"Character {args['name']} already exists.")
                else:
                    self.bios[char] = self.Bio.blank_bio(msg.author.id, args['name'])
                    await respond(msg, f"**AFFIRMATIVE. ANALYSIS: Created character {args['name']}.**")

            if not (self.bios[char].author == msg.author.id or args['dump'] or
                    self._bio.perms.check_optional_permissions("edit_others", msg.author, msg.channel)):
                raise UserPermissionError("Character belongs to another user.")

            # setting one field of the bio to a given value
            if args['set']:
                field, value = args['set'][0], ' '.join(args['set'][1:])
                try:
                    self.bios[char].set(field, value)
                    await self._update_bio_pin(msg.guild, char)
                    await respond(msg, f"**AFFIRMATIVE. {field.capitalize()} {'' if value else 're'}set.**")
                except ValueError as e:
                    raise CommandSyntaxError(f"Exceeded length of field {field.capitalize()}: {e} characters.")
                except KeyError as e:
                    raise CommandSyntaxError(e)

            self.storage_file.save()

            # compiling the bio into a json file for storage and editing
            if args['dump']:
                t_bio = asdict(self.bios[char])
                del t_bio['author']
                t_bio['fullname'] = t_bio['name']
                t_bio['name'] = char
                t_bio = json.dumps(t_bio, indent=2, ensure_ascii=False)
                async with msg.channel.typing():
                    await respond(msg, "**AFFIRMATIVE. Completed file upload.**",
                                  file=discord.File(BytesIO(bytes(t_bio, encoding="utf-8")), filename=char + ".json"))

            # changing the bio key in the storage dict, effectively renaming it
            if args['rename']:
                new_name = self.Bio._name(' '.join(args['rename'])).lower()
                if new_name in self.bios:
                    raise UserPermissionError(f"Character {new_name} already exists")

                self.bios[new_name] = self.bios[char]
                del self.bios[char]

                if char in self.config['pinned_bios']:
                    self.config['pinned_bios'][new_name] = self.config['pinned_bios'][char]
                    del self.config['pinned_bios'][char]

                await respond(msg, f"**AFFIRMATIVE. Character {char} can now be accessed as {new_name}.**")
                char = new_name
                self.storage_file.save()

            # deletes the specified bio.
            if args['delete']:
                del self.bios[char]
                if char in self.config['pinned_bios']:
                    bio_msg = await msg.guild.get_channel(self.config['pinned_bios_channel'])\
                        .fetch_message(self.config['pinned_bios'][char])
                    await bio_msg.delete()  # deleting the actual record happens in on_message_delete
                self.storage_file.save()
                await respond(msg, f"**AFFIRMATIVE. Character {char} has been deleted.**")

        else:
            await respond(msg, None,
                          embed=self.bios[char].embed(msg.guild, self.config['race_roles']))

    @Command("UploadBio",
             doc="Parses a JSON file or a JSON codeblock to update/create character bios.\n"
                 "See output of ",
             syntax="(attach file to the message, or put JSON contents into a code block following the command)",
             category="role_play")
    async def _upload_bio(self, msg: discord.Message):
        if msg.attachments:
            # there is a file uploaded with the message, grab and decode it.
            _file = BytesIO()

            await msg.attachments[0].save(_file)
            try:
                data = decode_json(_file.getvalue())
            except ValueError as e:
                self.logger.exception("Could not decode uploaded bio file!", exc_info=True)
                raise CommandSyntaxError(e)
            except Exception as e:
                raise CommandSyntaxError(f"Not a valid JSON file: {e}")
        else:
            # no file, let's see if user given us a code block.
            args = msg.content.split(None, 1)
            if len(args) == 1:
                raise CommandSyntaxError("File or code block required.")

            # This regexp searches for something in a codeblock, inside {} inclusive.
            # There can be any amount of space between the codeblock ticks and the figure brackets.
            data = re.search("```.*({.+}).*```", args[1], re.DOTALL)

            if data:
                try:
                    data = json.loads(data.group(1))
                except ValueError as e:
                    raise CommandSyntaxError(f"Not a valid JSON string: {e}")
            else:
                raise CommandSyntaxError("Not a valid JSON code block.")

        if "name" not in data:
            raise CommandSyntaxError("Not a valid character file: No name.")

        name = self.Bio._name(data['name']).lower()

        # 'fullname' key only exists for users' sake and must be dealt with specially
        try:
            data['name'] = data['fullname'] or data['name']
            del data['fullname']
        except KeyError:
            pass

        if name in self.bios and self.bios[name].author != msg.author.id:
            raise UserPermissionError("Character belongs to another user.")

        new_char = self.Bio.blank_bio(msg.author.id, data['name'])

        # the Bio.set() method includes all the checks for length that we may need, just gotta let it do its thing.
        for field, value in data.items():
            try:
                new_char.set(field, value)
            except ValueError as e:
                raise CommandSyntaxError(f"Exceeded length of field {field.capitalize()}: {e} characters.")
            except KeyError:
                continue

        # just for nicer output
        old = name in self.bios

        self.bios[name] = new_char
        self.storage_file.save()
        await self._update_bio_pin(msg.guild, name)

        await respond(msg, f"**AFFIRMATIVE. Character {new_char.name} was {'updated' if old else 'created'}.**")

    @Command("ReloadBios", "ReloadBio",
             doc="Administrative function that reloads the bios from the file.",
             category="role_play",
             bot_maintainers_only=True)
    async def _reload_bio(self, msg: discord.Message):
        self.storage_file.load()
        self.bios = self.storage["bios"]
        await respond(msg, "**AFFIRMATIVE. Bios reloaded from file.**")

    @Command("PinBio",
             doc="Generates an automatically updated bio post.\n"
                 "All such messages must be in the same channel, which is set by the first message.\n"
                 "To unpin a bio, simply delete the message.",
             syntax="(character)",
             perms={"manage_messages"},
             category="role_play",
             run_anywhere=True,
             delete_call=True)
    async def _pin_bio(self, msg):
        g_cfg = self.config

        if g_cfg.setdefault('pinned_bios_channel', msg.channel.id) != msg.channel.id:
            if g_cfg['pinned_bios']:
                raise CommandSyntaxError(f"Auto-pinned bios must all be in channel "
                                         f"<#{self.config['pinned_bios_channel']}>.")
            else:
                g_cfg['pinned_bios_channel'] = msg.channel.id

        try:
            char = self.Bio._name(msg.clean_content.split(None, 1)[1])
        except IndexError:
            raise CommandSyntaxError

        if char not in self.bios:
            raise CommandSyntaxError(f"No bio with id {char}.")

        if char in g_cfg.setdefault('pinned_bios', {}):
            raise CommandSyntaxError(f"Bio with id {char} is already pinned.")

        message = await respond(msg, None, embed=self.bios[char].embed(msg.guild, g_cfg['race_roles']))

        g_cfg['pinned_bios'][char] = message.id

    # util commands

    async def _update_bio_pin(self, guild: discord.Guild, char: str):
        g_cfg = self.config
        if char in g_cfg['pinned_bios']:
            bio_msg = await guild.get_channel(g_cfg['pinned_bios_channel']).fetch_message(g_cfg['pinned_bios'][char])
            await bio_msg.edit(embed=self.bios[char].embed(guild, g_cfg['race_roles']))

    async def on_message_delete(self, msg: discord.Message):
        g_cfg = self.config

        if 'pinned_bios' in g_cfg and msg.id in g_cfg['pinned_bios'].values():
            key = [k for k, v in g_cfg['pinned_bios'].items() if v == msg.id].pop()
            del g_cfg['pinned_bios'][key]



from red_star.plugin_manager import BasePlugin
from red_star.rs_utils import respond
from red_star.command_dispatcher import Command
from red_star.rs_errors import CommandSyntaxError
import discord
from io import BytesIO


class DumpChannel(BasePlugin):
    name = "dump_channel"
    version = "1.0"
    author = "GTG3000"
    description = "A plugin for dumping messages in a channel to a text file."

    @Command("Dump",
             doc="Dumps the messages between two specified messages into a text file, inclusively.",
             syntax="(latest message ID) (earliest message ID) [filename]",
             perms={"manage_messages"},
             run_anywhere=True)
    async def _dump(self, msg: discord.Message):
        args = msg.content.split(" ", 3)
        if len(args) < 3:
            raise CommandSyntaxError("Wrong number of arguments.")
        try:
            m_start = int(args[1])
        except ValueError:
            raise CommandSyntaxError("First Argument is not a valid integer.")

        try:
            m_end = int(args[2])
        except ValueError:
            raise CommandSyntaxError("Second Argument is not a valid integer.")
        try:
            m_start = await msg.channel.fetch_message(m_start)
        except discord.NotFound:
            raise CommandSyntaxError(f"No message with ID {m_start}")
        try:
            m_end = await msg.channel.fetch_message(m_end)
        except discord.NotFound:
            raise CommandSyntaxError(f"No message with ID {m_end}")

        if len(args) > 3:
            t_name = args[3]+".txt"
        else:
            t_name = str(msg.created_at.timestamp())+".txt"

        s = "%Y-%m-%d %H:%M:%S"

        t_list = [f"{str(m_end.author)} @ {str(m_end.created_at.strftime(s))}\n{m_end.clean_content}\n\n"]

        async for msg in msg.channel.history(before=m_start, after=m_end, oldest_first=True, limit=None):
            t_list.append(f"{str(msg.author)} @ {str(msg.created_at.strftime(s))}\n{msg.clean_content}\n\n")

        t_list.append(f"{str(m_start.author)} @ {str(m_start.created_at.strftime(s))}\n{m_start.clean_content}")

        t_msg = await respond(msg, f"**AFFIRMATIVE. Processing file {t_name}.**")
        async with msg.channel.typing():
            await respond(msg, "**AFFIRMATIVE. Completed file upload.**",
                          file=discord.File(BytesIO(bytes("".join(t_list), encoding="utf-8")), filename=t_name))
        await t_msg.delete()



import discord

from red_star.plugin_manager import BasePlugin
from red_star.rs_utils import respond
from red_star.command_dispatcher import Command
from red_star.rs_errors import CommandSyntaxError
from random import randint
from collections import defaultdict
from enum import Enum
import re


roll_tokens = re.compile(r"\d+(?:\.\d+)?|d(?:\d+|f)[da]?|[+\-/*()]")


class Adv(Enum):
    n = 0
    a = 1
    d = 2


def parse_roll(roll_string: str):
    rolls = []

    def rpn(string: str):
        # tokenize by only finding the valid operators
        tokens = roll_tokens.findall(string.lower())

        # turn into reverse polish notation for easier parsing
        prec = defaultdict(lambda: 3, {
            '(': 0,
            '-': 1,
            '+': 1,
            '*': 2,
            '/': 2
        })

        polish = []
        ops = []
        unary_minus_flag = True  # ugh
        new_expr_flag = False  # because people may not write right

        for token in tokens:
            try:
                try:
                    val = int(token)
                except ValueError:
                    val = float(token)

                if new_expr_flag:
                    polish.extend(ops[::-1])
                    ops = []

                polish.append(val)
                unary_minus_flag = False
                new_expr_flag = True
            except ValueError:
                new_expr_flag = False
                if token == '(':
                    ops.append(token)
                    unary_minus_flag = True
                elif token == ')':
                    while ops and ops[-1] != '(':
                        polish.append(ops.pop())
                    if ops:
                        ops.pop()
                    unary_minus_flag = False
                elif token[0] == 'd':
                    polish.append(token)
                    unary_minus_flag = False
                else:
                    # unary minuses are hard.
                    if unary_minus_flag and token == '-':
                        token = '_'
                    unary_minus_flag = token in '*/+-'

                    # as cool as it would be to have strings of data then strings of operators, this works better
                    while ops and prec[ops[-1]] >= prec[token]:
                        polish.append(ops.pop())
                    ops.append(token)

        polish.extend(ops[::-1])

        return polish

    def roll(dice: str, stack: list):
        num = min(max(round(stack.pop()), 1), 10000) if stack else 1

        if dice[-1] in 'ad':
            reroll = Adv.a if dice[-1] == 'a' else Adv.d
            side = 'f' if dice[1:-1] == 'f' else int(dice[1:-1])
        else:
            reroll = Adv.n
            side = 'f' if dice[1:] == 'f' else int(dice[1:])

        if side == 'f':
            def diceroll():
                return randint(-1, 1)
        else:
            side = min(max(side, 2), 10000)

            def diceroll():
                return randint(1, side)

        roll_a = [diceroll() for _ in range(num)]
        roll_b = [diceroll() for _ in range(num)]

        if reroll == Adv.a:
            roll_a = roll_a if sum(roll_a) > sum(roll_b) else roll_b
        elif reroll == Adv.d:
            roll_a = roll_a if sum(roll_a) < sum(roll_b) else roll_b

        rolls.append(f"{num}d{side}{'a' if reroll == Adv.a else ('d' if reroll == Adv.d else '')} - {sum(roll_a):2d}"
                     f" {roll_a}")

        stack.append(sum(roll_a))

    # evaluate using reverse polish notation rules

    polish = rpn(roll_string)
    stack = []

    u_ops = {
        '_': lambda: stack.append(-stack.pop())
    }

    b_ops = {
        '+': lambda x, y: stack.append(y + x),
        '-': lambda x, y: stack.append(y - x),
        '/': lambda x, y: stack.append(y / x),
        '*': lambda x, y: stack.append(y * x)
    }

    while polish:
        token = polish.pop(0)
        if isinstance(token, (int, float)):
            stack.append(token)
        elif len(stack) > 1 and token in b_ops:
            b_ops[token](stack.pop(), stack.pop())
        elif len(stack) > 0 and token in u_ops:
            u_ops[token]()
        elif token[0] == 'd':
            roll(token, stack)

    return stack, rolls


class DiceRoll(BasePlugin):
    name = "diceroll"
    description = "A plugin for rolling dice and dice accessories."
    version = "1.0"
    author = "GTG3000"

    @Command("Roll",
             doc="Rolls a specified amount of specified dice with specified bonus and advantage/disadvantage.\n\n"
                 "The parser can evaluate full dice expressions, including basic calculations such as addition, "
                 "subtraction, multiplication and division.\n"
                 "It is also capable of doing multiple rolls per same expression, '4df 4df' for example.\n"
                 "'dn' is treated as an unary operator, so for example '2d6 d6' will roll 2d6 and then roll that "
                 "amount of d6.",
             syntax="[number]D(die/F)[A/D][+/-bonus]",
             category="role_play",
             run_anywhere=True)
    async def _roll(self, msg: discord.Message):
        args = msg.clean_content.split(None, 1)
        if len(args) < 2:
            raise CommandSyntaxError("Requires a roll expression.")

        results, rolls = parse_roll(args[1])
        roll_args = ' '.join(roll_tokens.findall(args[1])).upper()

        results = (str(x) for x in results)

        if rolls:
            t_string = f"**ANALYSIS: {msg.author.display_name} has attempted a " \
                       f"{roll_args} roll, getting {', '.join(results)}.\n" \
                       f"ANALYSIS: Rolled dice:** ```\n"
            if len(rolls[0]) + len(t_string) <= 1996:
                for r in rolls:
                    if len(t_string) + len(r) > 1996:
                        t_string += r[:1993 - len(t_string)] + '...'
                        break
                    else:
                        t_string += r + '\n'
                t_string += '```'
            else:
                t_string += rolls[0][:1990 - len(t_string)] + '...\n```'
            await respond(msg, t_string)
        else:
            await respond(msg, f"**ANALYSIS: expression {roll_args} evaluated. Result: {', '.join(results)}**")
